#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""
pure_software.py

Description: Unittest for software not requiring hardware. 
Serial port required to open an HCI instance.

"""
import os
import unittest

from max_ble_hci import BleHci, utils
from max_ble_hci.hci_packets import EventCode, EventPacket, StatusCode


class Software(unittest.TestCase):
    """Softare Unit Test"""

    # pylint: disable=missing-function-docstring
    def hci_required(self):
        if os.environ.get("TEST_PORT"):
            port = os.environ.get("TEST_PORT")
        else:
            ports = utils.get_serial_ports()
            if not ports:
                return
            port = ports[0]

        hci = BleHci(port)

        expected_fips1 = [170, 187, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        expected_fips2 = [104, 101, 108, 108, 111, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        fips1 = hci.convert_fips197(0xAABB)
        fips2 = hci.convert_fips197("hello")
        self.assertEqual(fips1, expected_fips1)
        self.assertEqual(fips2, expected_fips2)

    def test_utils(self):
        ans = utils.to_le_nbyte_list(0xAABB, 4)
        self.assertEqual(ans, [0xBB, 0xAA, 0x00, 0x00])

        ans = utils.le_list_to_int(ans)
        self.assertEqual(ans, 0xAABB)

        good_data = [1, 2, 3, 4]
        ans = utils.can_represent_as_bytes(good_data)
        self.assertTrue(ans)

        bad_data = [1, 2009, 3, 4]
        ans = utils.can_represent_as_bytes(bad_data)
        self.assertFalse(ans)

        address = utils.convert_str_address("00:11:22:33:44:55")
        self.assertEqual(address, 0x001122334455)

        self.assertEqual(utils.byte_length(0), 1)
        self.assertEqual(utils.byte_length(1), 1)
        self.assertEqual(utils.byte_length(254), 1)
        self.assertEqual(utils.byte_length(30_000), 2)
        self.assertEqual(utils.byte_length(-1), 1)

        reset_command = "0e0401030c00"
        evt = EventPacket.from_bytes(bytes.fromhex(reset_command))
        self.assertEqual(evt.evt_code, EventCode.COMMAND_COMPLETE)
        self.assertEqual(evt.get_return_params(), StatusCode.SUCCESS.value)


if __name__ == "__main__":
    unittest.main()



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""Contains objects used for the creation of HCI packets."""
# pylint: disable=too-many-arguments
from __future__ import annotations

import warnings
from enum import Enum
from typing import List, Optional, Union
from .utils import byte_length
from .constants import Endian, PhyOption
from .packet_codes import EventCode, EventSubcode, StatusCode
from .packet_defs import (
    OCF,
    OGF,
    ControllerOCF,
    InformationalOCF,
    LEControllerOCF,
    LinkControlOCF,
    NOpOCF,
    PacketType,
    StatusOCF,
    VendorSpecificOCF,
)


class CommandPacket:
    """Serializer for HCI command packets.

    Object defines a container/serializer for HCI command
    packets. Initializing an instance of the object creates
    a container which stores the desired command opcode and
    parameters. A serialized command can then be generated
    through the use of the `to_bytes` function. In the event
    that an opcode is needed but a full packet is not, the
    static method `make_hci_opcode` can be called without
    initializing an instance of the object.

    Parameters
    ----------
    ogf : Union[OGF, int]
        Opcode group field.
    ocf : Union[OCF, int]
        Opcode command field.
    params : Union[List[int], int], optional
        Command parameters, if any.

    Attributes
    ----------
    ogf : OGF
        Opcode group field.
    ocf : OCF
        Opcode command field.
    length : int
        Total length of command parameters.
    opcode : int
        Command opcode.
    params : Union[List[int], int], optional
        Command parameters, if any.

    """

    def __init__(
        self,
        ogf: Union[OGF, int],
        ocf: Union[OCF, int],
        params: Optional[Union[List[int], int]] = None,
    ):
        self.ogf = self._enum_to_int(ogf)
        self.ocf = self._enum_to_int(ocf)
        self.length = self._get_length(params)
        self.opcode = CommandPacket.make_hci_opcode(self.ogf, self.ocf)
        if params is not None:
            self.params = params if isinstance(params, list) else [params]
        else:
            self.params = None

    def __repr__(self) -> str:
        ogf, ocf = CommandPacket.get_ogf_ocf(self.opcode)

        cmd_str = f"OGF: {ogf}\n"
        cmd_str += f"OCF: {ocf}\n"
        cmd_str += f"OPCODE: {hex(self.opcode)}\n"
        cmd_str += f"LENGTH: {self.length}\n"
        cmd_str += f"PARAMS: {self.params}"

        return cmd_str

    def _enum_to_int(self, num):
        """Convert an enumeration value to an integer.

        PRIVATE

        """
        if isinstance(num, Enum):
            return num.value
        return num

    def _get_length(self, params):
        """Get parameters length.

        PRIVATE

        """
        if params is None:
            return 0
        if isinstance(params, int):
            return byte_length(params)

        return sum(byte_length(x) for x in params)

    @staticmethod
    def get_ogf_ocf(opcode) -> tuple[OGF, Union[ControllerOCF, LEControllerOCF]]:
        """Get OGF and OCF from opcode of command packet

        Parameters
        ----------
        opcode : int
            2 Bytes opcode

        Returns
        -------
        tuple
            OGF, OCF
        """

        ogf = OGF(opcode >> 10)
        ocf = opcode & 0x03F

        if ogf == OGF.NOP:
            ocf = NOpOCF.NOP
        elif ogf == OGF.LINK_CONTROL:
            ocf = LinkControlOCF(ocf)

        elif ogf == OGF.CONTROLLER:
            ocf = ControllerOCF(ocf)
        elif ogf == OGF.INFORMATIONAL:
            ocf = InformationalOCF(ocf)
        elif ogf == OGF.STATUS:
            ocf = StatusOCF(ocf)
        elif ogf == OGF.LE_CONTROLLER:
            ocf = LEControllerOCF(ocf)

        elif ogf == OGF.VENDOR_SPEC:
            ocf = VendorSpecificOCF(ocf)
        else:
            ocf = None

        return ogf, ocf

    @staticmethod
    def from_bytes(command: bytearray) -> CommandPacket:
        """Convert command from byte array to command packet

        Parameters
        ----------
        command : bytearray
            raw command in bytes

        Returns
        -------
        CommandPacket
            Decoded command packet
        """
        opcode = command[1] << 8 | command[0]

        ogf, ocf = CommandPacket.get_ogf_ocf(opcode=opcode)

        ogf = OGF(ogf)
        # NO Link policy, testing OCF?

        if ogf == OGF.NOP:
            ocf = NOpOCF(ocf)
        elif ogf == OGF.LINK_CONTROL:
            ocf = LinkControlOCF(ocf)
        elif ogf == OGF.CONTROLLER:
            ocf = OCF.CONTROLLER(ocf)
        elif ogf == OGF.INFORMATIONAL:
            ocf = InformationalOCF(ocf)
        elif ogf == OGF.STATUS:
            ocf = StatusOCF(ocf)
        elif ogf == OGF.LE_CONTROLLER:
            ocf = LEControllerOCF(ocf)
        elif ogf == OGF.VENDOR_SPEC:
            ocf = VendorSpecificOCF(ocf)

        return CommandPacket(ogf=ogf, ocf=ocf, params=list(command[2:]))

    @staticmethod
    def make_hci_opcode(ogf: Union[OGF, int], ocf: Union[OCF, int]) -> int:
        """Make an HCI opcode.

        Creates an HCI opcode from the given Opcode Group Field (OGF)
        and Opcode Command Field (OCF) values.

        Parameters
        ----------
        ogf : Union[OGF, int]
            Opcode group field.
        ocf : Union[OCF, int]
            Opcode command field.

        Returns
        -------
        int
            The generated HCI opcode.

        """
        if not isinstance(ogf, int):
            if isinstance(ogf, Enum):
                ogf = ogf.value
            else:
                raise TypeError(
                    "Parameter 'ogf' must be an integer or an OGF enumeration."
                )

        if not isinstance(ocf, int):
            if isinstance(ocf, Enum):
                ocf = ocf.value
            else:
                raise TypeError(
                    "Parameter 'ogf' must be an integer or an OCF enumeration."
                )

        return (ogf << 10) | ocf

    def to_bytes(self, endianness: Endian = Endian.LITTLE) -> bytearray:
        """Serialize a command packet.

        Serializes a command packets from the stored attribute
        values into a command data byte array.

        Parameters
        ----------
        endianness : Endian
            Endian byte order to apply during serialization.

        Returns
        -------
        bytearray
            The serialized command.

        """
        serialized_cmd = bytearray()
        serialized_cmd.append(PacketType.COMMAND.value)
        serialized_cmd.append(self.opcode & 0xFF)
        serialized_cmd.append((self.opcode & 0xFF00) >> 8)

        serialized_cmd.append(self.length)

        if self.params is not None:
            for param in self.params:
                num_bytes = byte_length(param)
                try:
                    serialized_cmd.extend(param.to_bytes(num_bytes, endianness.value))
                except OverflowError:
                    serialized_cmd.extend(
                        param.to_bytes(num_bytes, endianness.value, signed=True)
                    )

        return serialized_cmd


class ExtendedPacket:
    """Serializer for HCI extended command packets.

    Object defines a container/serializer for HCI extended command
    packets. Initializing an instance of the object creates a container
    which stores the desired extended command opcode and payload. A
    serialized command can then be generated through the use of the
    `to_bytes` function. In the event that an opcode is needed but a
    full packet is not, the static method `make_hci_opcode` can be used
    without initializing an instance of the object.

    Parameters
    ----------
    ogf : Union[OGF, int]
        Opcode group field.
    ocf : Union[OCF, int]
        Opcode command field.
    payload : Union[List[int], int], optional
        Command parameters, if any.

    Attributes
    ----------
    ogf : OGF
        Opcode group field.
    ocf : OCF
        Opcode command field.
    length : int
        Total length of command parameters.
    opcode : int
        Command opcode.
    payload : Union[List[int], int], optional
        Command parameters, if any.

    """

    def __init__(
        self,
        ogf: Union[OGF, int],
        ocf: Union[OCF, int],
        payload: Optional[Union[List[int], int]] = None,
    ):
        self.ogf = self._enum_to_int(ogf)
        self.ocf = self._enum_to_int(ocf)
        self.length = self._get_length(payload)
        self.opcode = ExtendedPacket.make_hci_opcode(self.ogf, self.ocf)
        if payload is not None:
            self.payload = payload if isinstance(payload, list) else [payload]
        else:
            self.payload = None

    def __repr__(self):
        return str(self.__dict__)

    def _enum_to_int(self, num):
        """Convert an enumeration value to an integer.

        PRIVATE

        """
        if isinstance(num, Enum):
            return num.value
        return num

    def _get_length(self, pld):
        """Get payload length.

        PRIVATE

        """
        if pld is None:
            return 0
        if isinstance(pld, int):
            return byte_length(pld)

        return sum(byte_length(x) for x in pld)

    @staticmethod
    def make_hci_opcode(ogf: Union[OGF, int], ocf: Union[OCF, int]) -> int:
        """Make an HCI opcode.

        Creates an HCI opcode from the given Opcode Group Field (OGF)
        and Opcode Command Field (OCF) values.

        Parameters
        ----------
        ogf : Union[OGF, int]
            Opcode group field.
        ocf : Union[OCF, int]
            Opcode command field.

        Returns
        -------
        int
            The generated HCI opcode.

        """
        if not isinstance(ogf, int):
            if isinstance(ogf, Enum):
                ogf = ogf.value
            else:
                raise TypeError(
                    "Parameter 'ogf' must be an integer or an OGF enumeration."
                )

        if not isinstance(ocf, int):
            if isinstance(ocf, Enum):
                ocf = ocf.value
            else:
                raise TypeError(
                    "Parameter 'ogf' must be an integer or an OCF enumeration."
                )

        return (ogf << 10) | ocf

    def to_bytes(self, endianness: Endian = Endian.LITTLE) -> bytearray:
        """Serialize a command packet.

        Serializes a command packets from the stored attribute
        values into a command data byte array.

        Parameters
        ----------
        endianness : Endian
            Endian byte order to apply during serialization.

        Returns
        -------
        bytearray
            The serialized command.

        """
        serialized_cmd = bytearray()
        serialized_cmd.append(PacketType.EXTENDED.value)
        serialized_cmd.append(self.opcode & 0xFF)
        serialized_cmd.append((self.opcode & 0xFF00) >> 8)

        serialized_cmd.append(self.length & 0xFF)
        serialized_cmd.append((self.length & 0xFF00) >> 8)

        if self.payload is not None:
            for param in self.payload:
                num_bytes = byte_length(param)
                try:
                    serialized_cmd.extend(param.to_bytes(num_bytes, endianness.value))
                except OverflowError:
                    serialized_cmd.extend(
                        param.to_bytes(num_bytes, endianness.value, signed=True)
                    )

        return serialized_cmd


class AsyncPacket:
    """Deserializer for HCI ACL packets.

    Object defines a deserializer/data container for HCI
    Asynchronous Connection-Less packets. To create an
    instance directly from bytes, use the static function
    `from_bytes`.

    Parameters
    ----------
    handle : int
        Packet handle value.
    pb_flag : int
        Packet PB flag.
    bc_flag : int
        Packet BC flag.
    length : int
        Packet data length.
    data : bytes
        Packet data.

    Attributes
    ----------
    handle : int
        Packet handle value.
    pb_flag : int
        Packet PB flag.
    bc_flag : int
        Packet BC flag.
    length : int
        Packet data length.
    data : bytes
        Packet data.

    """

    def __init__(
        self, handle: int, pb_flag: int, bc_flag: int, length: int, data: bytes
    ):
        self.handle = handle
        self.pb_flag = pb_flag
        self.bc_flag = bc_flag
        self.length = length
        self.data = data

    def __repr__(self) -> str:
        return str(self.__dict__)

    @staticmethod
    def from_bytes(pkt: bytes) -> AsyncPacket:
        """Deserialize an HCI ACL packet.

        Deserializes an HCI Asynchronous Connection-Less packet
        from a bytes object.

        Parameters
        ----------
        pkt : bytes
            Serialized async packet.

        Returns
        -------
        AsyncPacket
            The deserialized packet.

        """
        return AsyncPacket(
            handle=(pkt[0] & 0xF0) + (pkt[1] << 8),
            pb_flag=(pkt[0] & 0xC) >> 2,
            bc_flag=pkt[0] & 0x3,
            length=pkt[2] + (pkt[3] << 8),
            data=pkt[4:] if pkt[4:] else None,
        )


class EventPacket:
    """Deserializer for HCI event packets.

    Object defines a deserializer/data container for HCI
    event packets. To create an instance directly from
    bytes, use the static function `from_bytes`. Event
    packet return parameters can be retrieved by calling
    the `get_return_params` function once an instance of
    the object has been created.

    Parameters
    ----------
    evt_code : int
        Packet event code.
    length : int
        Packet data length.
    status : int
        Packet status code.
    evt_params : bytes
        Packet return parameters.
    evt_subcode : int, optional
        Packet event subcode.

    Attributes
    ----------
    evt_code : EventCode
        Packet event code.
    length : int
        Packet data length.
    status : StatusCode
        Packet status code.
    evt_subcode : EventSubcode, optional
        Packet event subcode
    evt_params : bytes
        Packet return parameters.

    """

    def __init__(
        self,
        evt_code: int,
        length: int,
        status: int,
        evt_params: bytes,
        evt_subcode: Optional[int] = None,
    ):
        try:
            self.evt_code = EventCode(evt_code)
        except ValueError:
            warnings.warn(
                f"Unknown event code {evt_code}. Storing as byte object.",
                RuntimeWarning,
            )
            self.evt_code = evt_code
        self.length = length
        self.status = StatusCode(status) if status is not None else None
        self.evt_subcode = EventSubcode(evt_subcode) if evt_subcode else None

        self.evt_params = evt_params

    def __repr__(self):
        return str(self.__dict__)

    @staticmethod
    def from_bytes(serialized_event: bytes) -> EventPacket:
        """Deserialize an HCI event packet.

        Deserializes an HCI event packet from a bytes object.

        Parameters
        ----------
        serialized_event : bytes
            Serialized event packet.

        Returns
        -------
        EventPacket
            The deserialized packet.


        """
        if serialized_event[0] == EventCode.COMMAND_COMPLETE.value:
            pkt = EventPacket(
                evt_code=serialized_event[0],
                length=serialized_event[1],
                status=StatusCode(serialized_event[5]),
                evt_params=serialized_event[2:],
            )

        elif serialized_event[0] == EventCode.HARDWARE_ERROR.value:
            pkt = EventPacket(
                evt_code=serialized_event[0],
                length=serialized_event[1],
                status=StatusCode.ERROR_CODE_HW_FAILURE.value,
                evt_params=serialized_event[2:],
            )
        elif serialized_event[0] == EventCode.NUM_COMPLETED_PACKETS.value:
            pkt = EventPacket(
                evt_code=serialized_event[0],
                length=serialized_event[1],
                status=StatusCode.SUCCESS.value,
                evt_params=serialized_event[2:],
            )
        elif serialized_event[0] == EventCode.DATA_BUFF_OVERFLOW.value:
            pkt = EventPacket(
                evt_code=serialized_event[0],
                length=serialized_event[1],
                status=None,
                evt_params=serialized_event[2:],
            )
        elif serialized_event[0] == EventCode.LE_META.value:
            pkt = EventPacket(
                evt_code=serialized_event[0],
                length=serialized_event[1],
                status=None,
                evt_params=serialized_event[3:],
                evt_subcode=serialized_event[2],
            )
        elif serialized_event[0] == EventCode.AUTH_PAYLOAD_TIMEOUT_EXPIRED.value:
            pkt = EventPacket(
                evt_code=serialized_event[0],
                length=serialized_event[1],
                status=None,
                evt_params=serialized_event[2:],
            )
        elif serialized_event[0] == EventCode.VENDOR_SPEC:
            pkt = EventPacket(
                evt_code=serialized_event[0],
                length=serialized_event[1],
                status=serialized_event[2],
                evt_params=serialized_event[3:],
            )
        else:
            pkt = EventPacket(
                evt_code=serialized_event[0],
                length=serialized_event[1],
                status=serialized_event[2],
                evt_params=serialized_event[3:],
            )
        return pkt

    def get_return_params(
        self,
        param_lens: Optional[List[int]] = None,
        endianness: Endian = Endian.LITTLE,
        signed: bool = False,
    ) -> Union[List[int], int]:
        """Retrieve packet return parameters.

        Parses the packet return parameters from the bytes stored
        in the `evt_params` attribute in accordance with the given
        lengths and deserialization parameters.

        Parameters
        ----------
        param_lens : List[int], optional
            The length values of each expected return parameter. If
            only 1 return is expected, this value does not need to
            be provided.
        endianness : Endian
            Endian byte order to apply during deserialization.
        signed : bool
            Are the return values signed integers?

        Returns
        -------
        Union[List[int], int]
            The parsed return parameter(s).

        """
        # pylint: disable=unknown-option-value,possibly-used-before-assignment
        if self.evt_code == EventCode.COMMAND_COMPLETE:
            param_bytes = self.evt_params[4:]
        elif self.evt_subcode in (
            EventSubcode.CONNECTION_CMPLT,
            EventSubcode.PHY_UPDATE_CMPLT,
            EventSubcode.CONNECTION_UPDATE,
        ):
            param_bytes = self.evt_params
        else:
            param_bytes = self.evt_params

        if not param_lens:
            return int.from_bytes(param_bytes, endianness.value, signed=signed)

        if sum(param_lens) > len(param_bytes):
            raise ValueError(
                "Expected and actual number of return bytes do not match. "
                f"Expected={sum(param_lens)}, Actual={len(param_bytes)}"
            )

        return_params = []
        p_idx = 0
        # pylint: disable=unknown-option-value
        # pylint: disable=possibly-used-before-assignment
        for p_len in param_lens:
            return_params.append(
                int.from_bytes(param_bytes[p_idx : p_idx + p_len], endianness.value)
            )
            p_idx += p_len
        # pylint: enable=possibly-used-before-assignment
        # pylint: enable=unknown-option-value

        return return_params

    def _decode_le_meta(self):
        if self.evt_subcode == EventSubcode.CONNECTION_CMPLT:
            ret = self.get_return_params(param_lens=[1, 2, 1, 1, 6, 2, 2, 2, 1])

            return {
                "status": StatusCode(ret[0]),
                "handle": ret[1],
                "role": ret[2],
                "peer_addr_type": ret[3],
                "peer_addr": ret[4],
                "conn_interval": ret[5],
                "conn_latency": ret[6],
                "supervision_timeout": ret[7],
                "master_clock_accuracy": ret[8],
            }
        if self.evt_subcode == EventSubcode.PHY_UPDATE_CMPLT:
            ret = self.get_return_params([1, 2, 1, 1])
            return {
                "status": StatusCode(ret[0]),
                "handle": ret[1],
                "tx_phy": PhyOption(ret[2]),
                "rx_phy": PhyOption(ret[3]),
            }

        if self.evt_subcode == EventSubcode.CONNECTION_UPDATE:
            ret = self.get_return_params([1, 2, 2, 2, 2])
            return {
                "status": StatusCode(ret[0]),
                "handle": ret[1],
                "conn_interval": ret[2],
                "conn_latency": ret[3],
                "supervision_timeout": ret[4],
            }

        return None

    def decode(self) -> dict:
        """Decode parameters from EventPacket

        Returns
        -------
        dict
            Decoded event parameters
        """
        if self.evt_code == EventCode.LE_META:
            return self._decode_le_meta()

        return None



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""
hci_decode.py

Description: Basic decoding of raw HCI command packets

"""

import argparse
import binascii

# pylint: disable=redefined-builtin,import-error
from rich import print

from max_ble_hci.hci_packets import (
    OGF,
    AsyncPacket,
    CommandPacket,
    ControllerOCF,
    EventPacket,
    LEControllerOCF,
)
from max_ble_hci.packet_codes import EventMask, EventMaskLE
from max_ble_hci.packet_defs import PacketType

# pylint: enable=redefined-builtin,import-error


def _make_event_mask(params):
    event_mask = 0
    param_len = params[0]

    for i, value in enumerate(params[1:]):
        event_mask |= value << (param_len - (i + 1)) * 8
    return event_mask


def _parse_event_mask(command: CommandPacket):
    event_mask = _make_event_mask(command.params)

    # clear all rfu bits
    mask = EventMask.from_int(event_mask)
    mask_list = mask.as_str_list()

    print("[cyan]Event Mask\n-----------[/cyan]")
    for mask_str in mask_list:
        print(mask_str)


def _parse_event_mask_le(command: CommandPacket):
    event_mask = _make_event_mask(command.params)
    event_mask &= (1 << 34) | ((1 << 34) - 1)

    mask = EventMaskLE(event_mask)
    mask_list = mask.as_str_list()

    print("[cyan]Event Mask\n-----------[/cyan]")
    for mask_str in mask_list:
        print(mask_str)


def _parse_packet(command):
    packet_type = int(command[:2], 16)
    packet_type = PacketType(packet_type)

    command = binascii.unhexlify(command[2:])

    print(packet_type)
    if packet_type == PacketType.COMMAND:
        command = CommandPacket.from_bytes(command)
        print(command)
        ogf = OGF(command.ogf)
        if (
            ogf == OGF.LE_CONTROLLER
            and command.ocf == LEControllerOCF.SET_EVENT_MASK.value
        ):
            _parse_event_mask_le(command)

        elif (
            ogf == OGF.CONTROLLER and command.ocf == ControllerOCF.SET_EVENT_MASK.value
        ):
            _parse_event_mask(command)

    elif packet_type == PacketType.EXTENDED:
        pass
    elif packet_type == PacketType.ASYNC:
        print(AsyncPacket.from_bytes(command))
    elif packet_type == PacketType.EVENT:
        print(EventPacket.from_bytes(command))
    else:
        raise ValueError(f"Unnknown packet type {packet_type}")


def main():
    """MAIN"""
    parser = argparse.ArgumentParser(
        description="Evaluates connection sensitivity",
    )

    parser.add_argument("command", help="Central board")
    parser.add_argument("-e", "--event-mask", help="Command is event mask")

    args = parser.parse_args()

    command: str = args.command

    _parse_packet(command)


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""Contains definitions for BLE standard codes utilized in HCI packet creation/parsing."""
from __future__ import annotations

from enum import Enum, Flag
from typing import List, Union


class EventCode(Enum):
    """Supported HCI Event Codes"""

    DICON_COMPLETE = 0x05
    """Disconnection complete event."""

    ENC_CHANGE = 0x08
    """Encryption change event."""

    READ_REMOTE_VERSION_INFO_COMPLETE = 0x0C
    """Read remote version information complete event."""

    COMMAND_COMPLETE = 0x0E
    """Command complete event."""

    COMMAND_STATUS = 0x0F
    """Command status event."""

    HARDWARE_ERROR = 0x10
    """Hardware error event."""

    NUM_COMPLETED_PACKETS = 0x13
    """Number of completed packets event."""

    DATA_BUFF_OVERFLOW = 0x1A
    """Data buffer overflow event."""

    ENC_KEY_REFRESH_COMPLETE = 0x30
    """Encryption key refresh complete event."""

    LE_META = 0x3E
    """LE meta event."""

    AUTH_PAYLOAD_TIMEOUT_EXPIRED = 0x57
    """Authenticated payload timeout expired event."""

    VENDOR_SPEC = 0xFF
    """Vendor specific event."""


def _get_full_mask(mask_type: Union[EventMask, EventMaskLE]):
    full_mask = None
    for flag in mask_type:
        if full_mask is None:
            full_mask = flag
        full_mask |= flag

    return full_mask


class EventMask(Flag):
    """Controller Event Mask"""

    INQUIRY_COMPLETE = 1 << 0
    INQUIRY_RESULT = 1 << 1
    CONNECTION_COMPLETE = 1 << 2
    CONNECTION_REQUEST = 1 << 3
    DISCONNECTION_COMPLETE = 1 << 4
    AUTHENTICATION_COMPLETE = 1 << 5
    REMOTE_NAME_REQUEST_COMPLETE = 1 << 6
    ENCRYPTION_CHANGE = 1 << 7
    CHANGE_CONNECTION_LINK_KEY_COMPLETE = 1 << 8
    LINK_KEY_TYPE_CHANGED = 1 << 9
    READ_REMOTE_SUPPORTED_FEATURES_COMPLETE = 1 << 10
    READ_REMOTE_VERSION_INFORMATION_COMPLETE = 1 << 11
    QOS_SETUP_COMPLETE = 1 << 12
    HARDWARE_ERROR = 1 << 15
    FLUSH_OCCURRED = 1 << 16
    ROLE_CHANGE = 1 << 17
    MODE_CHANGE = 1 << 19
    RETURN_LINK_KEYS = 1 << 20
    PIN_CODE_REQUEST = 1 << 21
    LINK_KEY_REQUEST = 1 << 22
    LINK_KEY_NOTIFICATION = 1 << 23
    LOOPBACK_COMMAND = 1 << 24
    DATA_BUFFER_OVERFLOW = 1 << 25
    MAX_SLOTS_CHANGE = 1 << 26
    READ_CLOCK_OFFSET_COMPLETE = 1 << 27
    CONNECTION_PACKET_TYPE_CHANGED = 1 << 28
    QOS_VIOLATION = 1 << 29
    PREVIOUSLY_USED = 1 << 30
    PAGE_SCAN_REPETITION_MODE_CHANGE = 1 << 31
    FLOW_SPECIFICATION_COMPLETE = 1 << 32
    INQUIRY_RESULT_WITH_RSSI = 1 << 33
    READ_REMOTE_EXTENDED_FEATURES_COMPLETE = 1 << 34
    SYNCHRONOUS_CONNECTION_COMPLETE = 1 << 43
    SYNCHRONOUS_CONNECTION_CHANGED = 1 << 44
    SNIFF_SUBRATING = 1 << 45
    EXTENDED_INQUIRY_RESULT = 1 << 46
    ENCRYPTION_KEY_REFRESH_COMPLETE = 1 << 47
    IO_CAPABILITY_REQUEST = 1 << 48
    IO_CAPABILITY_RESPONSE = 1 << 49
    USER_CONFIRMATION_REQUEST = 1 << 50
    USER_PASSKEY_REQUEST = 1 << 51
    REMOTE_OOB_DATA_REQUEST = 1 << 52
    SIMPLE_PAIRING_COMPLETE = 1 << 53
    LINK_SUPERVISION_TIMEOUT_CHANGED = 1 << 55
    ENHANCED_FLUSH_COMPLETE = 1 << 56
    USER_PASSKEY_NOTIFICATION = 1 << 58
    KEYPRESS_NOTIFICATION = 1 << 59
    REMOTE_HOST_SUPPORTED_FEATURES_NOTIFICATION = 1 << 60
    LE_META = 1 << 61

    @staticmethod
    def get_full_mask():
        """Get full event mask

        Returns
        -------
        EventMask
            All events in event mask
        """
        return _get_full_mask(EventMask)

    @staticmethod
    def from_int(mask: int) -> EventMask:
        """From integer mask to EventMask class

        Parameters
        ----------
        mask : int
            Interger mask.

        Returns
        -------
        EventMask
            EventMask class
        """
        clear_mask = EventMask.get_full_mask()
        return EventMask(mask & clear_mask.value)

    def as_str_list(self) -> List[str]:
        """Get list of event mask strings

        Returns
        -------
        List[str]
            List of event mask strings
        """
        flags = [flag.name for flag in EventMask if flag in self]
        return flags


class EventMaskPage2(Flag):
    """Controller Event Mask"""

    NUM_CMPLT_DAT_BLOCKS = 1 << 8
    TRIG_CLK_CAPTURE = 1 << 14
    SYNC_TRAIN_CMPLT = 1 << 15
    SYNC_TRAIN_RECEIVED = 1 << 16
    CONNECTIONLESS_PERIPH_BROADCAST = 1 << 17
    CONNECTIONLESS_PERIPH_TIMEOUT = 1 << 18
    TRUNCATED_PAGE_CMPLT = 1 << 19
    PERIPH_PAGE_RESP_TIMEOUT = 1 << 20
    CONNECTIONLESS_PERIPH_BROADCAST_CH_MAP_CHANGE = 1 << 21
    INQUIRY_RESP_NOTIF = 1 << 22
    AUTH_PAYLOAD_TIMEOUT_EXPIRED = 1 << 23
    SAM_STATUS_CHANGED = 1 << 24
    ENC_CHANGED = 1 << 25

    @staticmethod
    def get_full_mask() -> EventMaskPage2:
        """Get full event mask page 2

        Returns
        -------
        EventMaskPage2
            Full event mask page 2
        """
        return _get_full_mask(EventMaskPage2)

    @staticmethod
    def from_int(mask: int) -> EventMaskPage2:
        """From integer mask to EventMask class

        Parameters
        ----------
        mask : int
            Interger mask.

        Returns
        -------
        EventMask
            EventMask class
        """
        clear_mask = EventMask.get_full_mask()
        return EventMask(mask & clear_mask.value)

    def as_str_list(self) -> List[str]:
        """Get list of event mask strings

        Returns
        -------
        List[str]
            List of event mask strings
        """
        flags = [flag.name for flag in EventMaskPage2 if flag in self]
        return flags


class EventMaskLE(Flag):
    """LE Event Mask"""

    CONNECTION_COMPLETE = 1 << 0
    ADV_REPORT = 1 << 1
    CONN_UPDATE_COMPLETE = 1 << 2
    READ_REMOTE_FEAT_CMPLT = 1 << 3
    LTK_REQUEST = 1 << 4
    REMOTE_CONN_PARAM_REQ = 1 << 5
    DATA_LEN_CHANGE = 1 << 6
    READ_LOCAL_P256_PUB_KEY_CMPLT = 1 << 7
    GEN_DHKEY_CMPLT = 1 << 8
    ENHANCED_CONNECTION_COMPLETE = 1 << 9
    DIRECTED_ADV_REPORT = 1 << 10
    PHY_UPDATE_COMPLETE = 1 << 11
    EXTENDED_ADVERTISING_REPORT = 1 << 12
    PERIODIC_ADV_SYNC_EST = 1 << 13
    PERIODIC_ADV_REPORT = 1 << 14
    PERIODIC_ADV_SYNC_LOST = 1 << 15
    SCAN_TIMEOUT = 1 << 16
    ADVERTISING_SET_TERMINATED = 1 << 17
    SCAN_REQUEST_RECEIVED = 1 << 18
    CHANNEL_SELECTION_ALGORITHM = 1 << 19
    CONNECTIONLESS_IQ_REPORT = 1 << 20
    CONNECTION_IQ_REPORT = 1 << 21
    CTE_REQUEST_FAILED = 1 << 22
    PERIODIC_ADV_SYNC_XFER_LOST = 1 << 23
    CIS_ESTABLISHED = 1 << 24
    CIS_REQUEST = 1 << 25
    CREATE_BIG_COMPLETE = 1 << 26
    TERMINATE_BIG_COMPLETE = 1 << 27
    BIG_SYNC_ESTABLISHED = 1 << 28
    BIG_SYNC_LOST = 1 << 29
    REQUEST_PEER_SCA_COMPLETE = 1 << 30
    PATH_LOSS_THRESHOLD = 1 << 31
    TX_POWER_REPORTING = 1 << 32
    BIG_INFO_ADV_REPORT = 1 << 33
    SUBRATE_CHANGE = 1 << 34

    @staticmethod
    def get_full_mask() -> EventMaskLE:
        """Get full LE Event mask

        Returns
        -------
        EventMaskLE
            Event mask with all events
        """
        return _get_full_mask(EventMaskLE)

    def as_str_list(self) -> List[str]:
        """List of flags as strings

        Returns
        -------
        List[str]
            List of flags as string
        """
        flags = [flag.name for flag in EventMaskLE if flag in self]
        return flags

    @staticmethod
    def from_int(mask: int) -> EventMaskLE:
        """From int to EventMaskLE class

        Parameters
        ----------
        mask : int
            Integer mask

        Returns
        -------
        EventMaskLE
            EventMaskLE class
        """
        clear_mask = EventMaskLE.get_full_mask()
        return EventMaskLE(mask & clear_mask.value)


class EventSubcode(Enum):
    """Supported LE Meta event subcodes."""

    CONNECTION_CMPLT = 0x1
    """Connection complete event."""

    ADVERTISING_REPORT = 0x2
    """Advertising report event."""

    CONNECTION_UPDATE = 0x3
    """Connection update complete event."""

    READ_REMOTE_FEATURES_COMPLETE = 0x4
    """Read remote feature complete event."""

    LTK_REQUEST = 0x5
    """Long term key request event."""

    REMOTE_CONNECTION_PARAMETER_REQ = 0x6
    """Remote connection parameter request event."""

    DATA_LENGTH_CHANGE = 0x7
    """Data length change event."""

    READ_LOCAL_P256_PUB_KEY_CMPLT = 0x8
    """Read local P-256 public key complete event."""

    GENERATE_DHKEY_CMPLT = 0x9
    """Generate DHKey complete event."""

    ENHANCED_CONNECTION_CMPLT = 0xA
    """Enhanced connection complete event."""

    DIRECTED_ADVERTISIING_REPORT = 0xB
    """Directed advertising report event."""

    PHY_UPDATE_CMPLT = 0xC
    """PHY update complete event."""

    EXTENDED_ADVERTISING_REPORT = 0xD
    """Extended advertising report event."""

    PERIODIC_ADVERTISING_SYNC_ESTABLISHED = 0xE
    """Periodic advertising sync established event."""

    PERIODIC_ADVERTISING_REPORT = 0xF
    """Periodic advertising report event."""

    PERIODIC_ADVERTISING_SYNC_LOST = 0x10
    """Periodic advertising sync lost event."""

    SCAN_TIMEOUT = 0x11
    """Scan timeout event."""

    ADVERTISING_SET_TERMINATED = 0x12
    """Advertising set terminated event."""

    SCAN_REQUEST_RECEIVED = 0x13
    """Scan request received event."""

    CHANNEL_SELECTION_ALGORITHM = 0x14
    """Channel selection algorithm event."""

    CONNECTIONLESS_IQ_REPORT = 0x15
    """Connectionless IQ report event."""

    CONNECTION_IQ_REPORT = 0x16
    """Connection IQ report event."""

    CTE_REQUEST_FAILED = 0x17
    """CTE request failed event."""

    PERIODIC_ADVERTISING_SYNC_TRANSER_RECEIVED = 0x18
    """Periodic advertising sync transfer received event."""

    CIS_ESTABLISHED = 0x19
    """CIS established event."""

    CIS_REQUEST = 0x1A
    """CIS request event."""

    CREATE_BIG_CMPLT = 0x1B
    """Create BIG complete event."""

    TERMINATE_BIG_CMPLT = 0x1C
    """Terminate BIG complete event."""

    BIG_SYNC_EST = 0x1D
    """BIG sync established event."""

    BIG_SYNC_LOST = 0x1E
    """BIG sync list event."""

    REQUEST_PEER_SCA_CMPLT = 0x1F
    """Request peeer SCA complete event."""

    PATH_LOSS_THRESHOLD = 0x20
    """Path loss threshold event."""

    TRANSMIT_POWER_REPORTING = 0x21
    """Transmit power reporting event."""

    BIGINFO_ADVERTISING_REPORT = 0x22
    """BIGInfo advertising report event."""


class StatusCode(Enum):
    """BLE-defined status codes."""

    SUCCESS = 0x00
    """Success."""

    ERROR_CODE_UNKNOWN_HCI_CMD = 0x01
    """Unknown HCI command error."""

    ERROR_CODE_UNKNOWN_CONN_ID = 0x2
    """Unknown connection identifier."""

    ERROR_CODE_HW_FAILURE = 0x03
    """Hardware failure."""

    ERROR_CODE_PAGE_TIMEOUT = 0x04
    """Page timeout."""

    ERROR_CODE_AUTH_FAILURE = 0x05
    """Authentication failure."""

    ERROR_CODE_PIN_KEY_MISSING = 0x06
    """PIN or key missing."""

    ERROR_CODE_MEM_CAP_EXCEEDED = 0x07
    """Memory capacity exceeded."""

    ERROR_CODE_CONN_TIMEOUT = 0x08
    """Connection timeout."""

    ERROR_CODE_CONN_LIMIT_EXCEEDED = 0x09
    """Connection limit exceeded."""

    ERROR_CODE_SYNCH_CONN_LIMIT_EXCEEDED = 0x0A
    """Synchronous connection limit to a device exceeded."""

    ERROR_CODE_ACL_CONN_ALREADY_EXISTS = 0x0B
    """Connection already exists."""

    ERROR_CODE_CMD_DISALLOWED = 0x0C
    """Command disallowed."""

    ERROR_CODE_CONN_REJ_LIMITED_RESOURCES = 0x0D
    """connection rejection due to limited resources."""

    ERROR_CODE_CONN_REJECTED_SECURITY_REASONS = 0x0E
    """Connection rejected due to security reasons."""

    ERROR_CODE_CONN_REJECTED_UNACCEPTABLE_BDADDR = 0x0F
    """Connection rejected dur to unacceptable BD address."""

    ERROR_CODE_CONN_ACCEPT_TIMEOUT_EXCEEDED = 0x10
    """Connection accept timeout exceeded."""

    ERROR_CODE_UNSUPPORTED_FEATURE_PARAM_VALUE = 0x11
    """Unsupported feature or parameter value."""

    ERROR_CODE_INVALID_HCI_CMD_PARAMS = 0x12
    """Invalid HCI command parameters."""

    ERROR_CODE_REMOTE_USER_TERM_CONN = 0x13
    """Remote user terminated connection."""

    ERROR_CODE_REMOTE_DEVICE_TERM_CONN_LOW_RESOURCES = 0x14
    """Remote device terminated connection due to low resources."""

    ERROR_CODE_REMOTE_DEVICE_TERM_CONN_POWER_OFF = 0x15
    """Remote device terminated connection due to power off."""

    ERROR_CODE_CONN_TERM_BY_LOCAL_HOST = 0x16
    """Connection terminated by local host."""

    ERROR_CODE_REPEATED_ATTEMPTS = 0x17
    """Repeated attempts."""

    ERROR_CODE_PAIRING_NOT_ALLOWED = 0x18
    """Pairing not allowed."""

    ERROR_CODE_UNKNOWN_LMP_PDU = 0x19
    """Unknown LMP PDU."""

    ERROR_CODE_UNSUPPORTED_REMOTE_FEATURE = 0x1A
    """Unsupported remote feature."""

    ERROR_CODE_SCO_OFFSET_REJ = 0x1B
    """SCO offset rejected."""

    ERROR_CODE_SCO_INTERVAL_REJ = 0x1C
    """SCO interval rejected."""

    ERROR_CODE_SCO_AIR_MODE_REJ = 0x1D
    """SCO air mode rejected."""

    ERROR_CODE_INVALID_LMP_PARAMS = 0x1E
    """Invalid LMP parameters / Invalid LL parameters."""

    ERROR_CODE_UNSPECIFIED_ERROR = 0x1F
    """Unspecified error."""

    ERROR_CODE_UNSUPPORTED_LMP_PARAM_VAL = 0x20
    """Unsupported LMP parameter value / Unsupported LL parameter value."""

    ERROR_CODE_ROLE_CHANGE_NOT_ALLOWED = 0x21
    """Role change not allowed."""

    ERROR_CODE_LMP_LL_RESP_TIMEOUT = 0x22
    """LMP response timeout / LL response timeout."""

    ERROR_CODE_LMP_ERR_TRANSACTION_COLLISION = 0x23
    """LMP error transaction collision / LL procedure collision."""

    ERROR_CODE_LMP_PDU_NOT_ALLOWED = 0x24
    """LMP PDU not allowed."""

    ERROR_CODE_ENCRYPT_MODE_NOT_ACCEPTABLE = 0x25
    """Encryption mode not acceptable."""

    ERROR_CODE_LINK_KEY_CAN_NOT_BE_CHANGED = 0x26
    """Link key cannot be changed."""

    ERROR_CODE_REQ_QOS_NOT_SUPPORTED = 0x27
    """Requested QoS not supported."""

    ERROR_CODE_INSTANT_PASSED = 0x28
    """Instance passed."""

    ERROR_CODE_PAIRING_WITH_UNIT_KEY_NOT_SUPPORTED = 0x29
    """Pairing with unit key not supported."""

    ERROR_CODE_DIFFERENT_TRANSACTION_COLLISION = 0x2A
    """Different transaction collision."""

    ERROR_CODE_RESERVED1 = 0x2B
    """Reserved for future use."""

    ERROR_CODE_QOS_UNACCEPTABLE_PARAM = 0x2C
    """QoS unacceptable parameter."""

    ERROR_CODE_QOS_REJ = 0x2D
    """QoS rejected."""

    ERROR_CODE_CHAN_ASSESSMENT_NOT_SUPPORTED = 0x2E
    """Channel classification not supported error."""

    ERROR_CODE_INSUFFICIENT_SECURITY = 0x2F
    """Insufficient security."""

    ERROR_CODE_PARAM_OUT_OF_MANDATORY_RANGE = 0x30
    """Parameter out of mandatory range."""

    ERROR_CODE_RESERVED2 = 0x31
    """Reserved for future use."""

    ERROR_CODE_ROLE_SWITCH_PENDING = 0x32
    """Role switch pending."""

    ERROR_CODE_RESERVED3 = 0x33
    """Reserved for future use."""

    ERROR_CODE_RESERVED_SLOT_VIOLATION = 0x34
    """Reserved slot violation."""

    ERROR_CODE_ROLE_SWITCH_FAILED = 0x35
    """Role switch failed."""

    ERROR_CODE_EXTENDED_INQUIRY_RESP_TOO_LARGE = 0x36
    """Extended inquiry response too large."""

    ERROR_CODE_SIMPLE_PAIRING_NOT_SUPPORTED_BY_HOST = 0x37
    """Secure simple pairing not supported by host."""

    ERROR_CODE_HOST_BUSY_PAIRING = 0x38
    """Host busy - pairing."""

    ERROR_CODE_CONN_REJ_NO_SUITABLE_CHAN_FOUND = 0x39
    """Connection rejected dur to no suitable channel found."""

    ERROR_CODE_CONTROLLER_BUSY = 0x3A
    """Controller busy."""

    ERROR_CODE_UNACCEPTABLE_CONN_INTERVAL = 0x3B
    """Unacceptable connection parameters."""

    ERROR_CODE_ADV_TIMEOUT = 0x3C
    """Advertising timeout."""

    ERROR_CODE_CONN_TERM_MIC_FAILURE = 0x3D
    """Connection terminated due to MIC failure."""

    ERROR_CODE_CONN_FAILED_TO_ESTABLISH = 0x3E
    """Connection failed to be established / Synchronization timeout."""

    ERROR_CODE_MAC_CONN_FAILED = 0x3F
    """MAC connection failed [previously used]."""

    ERROR_CODE_COARSE_CLK_ADJ_REJ = 0x40
    """
    Coarse cloack adjustment rejected but will try to adjust clock
    using clock dragging.
    """

    ERROR_CODE_TYPE0_SUBMAP_NOT_DEF = 0x41
    """Type0 submap noot defined."""

    ERROR_CODE_UNKNOWN_ADV_ID = 0x42
    """Unknown advertising identifier."""

    ERROR_CODE_LIMIT_REACHED = 0x43
    """Limit reached."""

    ERROR_CODE_OP_CANCELLED_BY_HOST = 0x44
    """Operation cancelled by host."""

    ERROR_CODE_PKT_TOO_LONG = 0x45
    """Packet too long."""

    DECODE_FAILURE = 0xFF
    """ADI vendor specific, returns when the status is not properly set."""



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""
utils.py

Description: Basic utilities to help withe BLE-HCI

"""

import glob
import os
import sys
from typing import List
import serial

DEFAULT_BAUDRATE = 115200


def get_serial_ports() -> List[str]:
    """Lists serial port names

    :raises EnvironmentError:
        On unsupported or unknown platforms
    :returns:
        A list of the serial ports available on the system
    """
    if sys.platform.startswith("win"):
        ports = [f"COM{(i + 1)}" for i in range(256)]
    elif sys.platform.startswith("linux") or sys.platform.startswith("cygwin"):
        # this excludes your current terminal "/dev/tty"
        ports = glob.glob("/dev/tty[A-Za-z]*")
    elif sys.platform.startswith("darwin"):
        ports = glob.glob("/dev/tty.*")
    else:
        raise EnvironmentError("Unsupported platform")

    result = []
    for port in ports:
        try:
            possible_port = serial.Serial(port)
            possible_port.close()
            result.append(port)
        except (OSError, serial.SerialException):
            pass

    serial_list_linux = "/dev/serial/by-id"

    if os.path.exists(serial_list_linux):
        dir_list = os.listdir(serial_list_linux)

        for i, file in enumerate(dir_list):
            dir_list[i] = os.path.join(serial_list_linux, file)
        if dir_list:
            result.extend(dir_list)

    return result


def to_le_nbyte_list(value: int, n_bytes: int) -> List[int]:
    """Create a list of little-endian bytes.

    Converts a multi-byte number into a list of single-byte
    values. The list is little endian.

    Parameters
    ----------
    value : int
        The multi-byte value that should be converted.
    n_bytes : int
        The expected byte length of the given value

    Returns
    -------
    List[int]
        The given value represented as a little endian
        list of single-byte values. The length is
        equivalent to the `n_bytes` parameter.

    """
    little_endian = []
    for i in range(n_bytes):
        num_masked = (value & (0xFF << 8 * i)) >> (8 * i)
        little_endian.append(num_masked)
    return little_endian


def le_list_to_int(nums: List[int]) -> int:
    """Create an integer from a little-endian list.

    Converts a little-endian list of single byte values
    to a single multi-byte integer.

    Parameters
    ----------
    nums : List[int]
        List containing single-byte values in little endian
        byte order.

    Returns
    -------
    int
        The multi-byte value created from the given list.

    """
    full_num = 0
    for i, num in enumerate(nums):
        full_num |= num << 8 * i

    return full_num


def can_represent_as_bytes(data: List[int]) -> bool:
    """Check whether all data can be represented by 1 byte

    Parameters
    ----------
    data : list
        data to check

    Returns
    -------
    bool
        True if all data is 1 byte or less
    """
    for item in data:
        if int(item) > 255:
            return False

    return True


def convert_str_address(addr: str) -> int:
    """Convert address as string to integer usable by HCI

    Parameters
    ----------
    addr : str
        Address. Ex: 00:11:22:33:44:55

    Returns
    -------
    int
        Address converted to integer value
    """
    return int(addr.replace(":", ""), 16)


def byte_length(data: int) -> int:
    """Get number of bytes needed to represent an integer value

    Parameters
    ----------
    data : int
        Number to get bytes needed

    Returns
    -------
    int
        Bytes needed to represent nearest value
    """

    if data == 0:
        return 1

    return (data.bit_length() + 7) // 8



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""
ad_types

Description: Advertising data types

"""

from enum import Enum


class ADTypes(Enum):
    """Advertising data types"""

    FLAGS = 0x1
    INCOMPLETE_16B_SC_UUID = 0x2
    COMPLETE_16B_SC_UUID = 0x3
    INCOMPLETE_32B_SC_UUID = 0x4
    COMPLETE_32B_SC_UUID = 0x5
    INCOMPLETE_128B_SC_UUID = 0x6
    COMPLETE_128B_SC_UUID = 0x7
    LOCAL_NAME_SHORT = 0x8
    LOCAL_NAME_COMPLETE = 0x9
    TX_POWER_LEVEL = 0x0A
    DEV_CLASS = 0x0D
    SIMPLE_PAIR_HASH_C192 = 0x0E
    SIMPLE_PAIR_RAND_R192 = 0x0F
    DEV_ID = 0x10
    SEC_MAN_TK_VALUE = 0x10
    SEC_MAN_OOB_FLAGS = 0x11
    PERIPH_CONN_INTERVAL_RANGE = 0x12
    SERV_SOLICITATION_16B = 0x14
    SERV_SOLICITATION_128B = 0x15
    SERV_DATA_16B = 0x16
    PUB_TARGET_ADDR = 0x17
    RAND_TARGET_ADDR = 0x18
    APPEARANCE = 0x19
    ADV_INTERVAL = 0x1A
    LE_BT_DEV_ADDR = 0x1B
    LE_ROLE = 0x1C
    SIMPLE_PAIR_HASH_C256 = 0x1D
    SIMPLE_PAIR_RAND_R256 = 0x1E
    SERV_SOLICITATION_32B = 0x1F
    SERV_DATA = 0x20
    LE_SEC_CONN_CONF_VALUE = 0x22
    LE_SEC_CONN_RAND_VALUE = 0x23
    URI = 0x24
    INDOOR_POSITIONING = 0x25
    TRANSPORT_DISC_DATA = 0x26
    LE_SUPP_FEAT = 0x27
    CHAN_MAP_UPDATE_IND = 0x28
    PB_ADV = 0x29
    MESH_MESSAGE = 0x2A
    MESH_BEACON = 0x2B
    BIG_INFO = 0x2C
    BROADCAST_CODE = 0x2D
    RESOLV_SET_ID = 0x2E
    ADV_INTERVAL_LONG = 0x2F
    BROADCAST_NAME = 0x30
    ENC_ADV_DATA = 0x31
    PER_ADC_RESP_TIME_INFO = 0x32
    ELECTRONIC_SHELF_LABEL = 0x34
    INFO_DATA_3D = 0x3D
    MANUFACTURER_SPEC_DATA = 0xFF



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""
max_ble_hci_cli.py

Description: CLI Client to use MAX-BLE-HCI

"""
import argparse
import logging
import os
import signal

# pylint: disable=unused-import,too-many-lines
try:
    import readline
except ImportError:
    import gnureadline as readline

# pylint: enable=unused-import
import sys
from argparse import RawTextHelpFormatter

# pylint: disable=import-error
from colorlog import ColoredFormatter

from max_ble_hci import BleHci
from max_ble_hci.constants import PayloadOption, PhyOption
from max_ble_hci.data_params import AdvParams, EstablishConnParams, ScanParams
from max_ble_hci.utils import convert_str_address

# pylint: enable=import-error


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)

formatter = ColoredFormatter(
    "%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(message)s",
    datefmt=None,
    reset=True,
    log_colors={
        "DEBUG": "cyan",
        "INFO": "green",
        "WARNING": "yellow",
        "ERROR": "red",
        "CRITICAL": "red,bg_white",
    },
    secondary_log_colors={},
    style="%",
)

console_handler.setFormatter(formatter)
logger.addHandler(console_handler)


EXIT_FUNC_MAGIC = 999
DEFAULT_BAUD = 115200
DEFAULT_ADV_INTERVAL = 0x60
DEFAULT_SCAN_INTERVAL = 0x10
DEFAULT_CONN_INTERVAL = 0x6  # 7.5 ms
DEFAULT_SUP_TIMEOUT = 0x64  # 1 s


class ArgumentParser(argparse.ArgumentParser):
    """Argument  Parser

    Parameters
    ----------
    argparse : base parser
        Child class adding functionality to basic Argument Parser
    """

    def error(self, message):  # pylint: disable=unused-argument
        logger.error("Invalid input. Refer to 'help' for available commands.'")
        self.exit(2)


def _hex_int(hex_str: str) -> int:
    return int(hex_str, 16)


def _signal_handler(_signal, _fname):  # pylint: disable=unused-argument
    print()
    sys.exit(0)


def _run_input_cmds(commands, terminal):
    for cmd in commands:
        try:
            # pylint: disable=used-before-assignment
            _args = terminal.parse_args(cmd.split())
            _args.func(_args)
        except AttributeError:
            continue
        # Catch SystemExit, allows user to ctrl-c to quit the current command
        except SystemExit as _err:
            if _err.code != 0:
                if _err.code == EXIT_FUNC_MAGIC:
                    sys.exit(0)

                sys.exit(_err.code)
    return True


def main():
    """
    MAIN
    """
    # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    command_state = ""

    # Setup the signal handler to catch the ctrl-C
    signal.signal(signal.SIGINT, _signal_handler)

    # Setup the command line description text
    cli_description = f"""
        Bluetooth Low Energy HCI tool.
        This tool is used in tandem with the BLE controller examples. 
        This tool sends HCI commands through the serial port to the target device. 
        It will receive and print the HCI events received from the target device.
        Serial port is configured as 8N1, no flow control, default baud rate of {str(DEFAULT_BAUD)}
        """

    # Parse the command line arguments
    parser = argparse.ArgumentParser(
        description=cli_description, formatter_class=RawTextHelpFormatter
    )

    parser.add_argument("--version", action="version", version="%(prog)s 1.2.0")

    parser.add_argument("serial_port", help="Serial port path or COM#")
    parser.add_argument(
        "-b",
        "--baud",
        dest="baudRate",
        type=int,
        default=DEFAULT_BAUD,
        help="Serial port baud rate. Default: " + str(DEFAULT_BAUD),
    )

    parser.add_argument(
        "-efc",
        "--enable-flow-control",
        action="store_false",
        default=False,
        help="Enable flow control Default: False",
    )
    parser.add_argument(
        "-m",
        "--monitor-trace-port",
        dest="monPort",
        default=None,
        help="Monitor Trace Msg Serial Port path or COM#. Default: None",
    )
    parser.add_argument(
        "-i",
        "--id-tag",
        dest="idtag",
        default="DUT",
        help="Board ID tag for printing trace messages. Default: None",
    )
    parser.add_argument(
        "-c",
        "--commands",
        dest="commands",
        nargs="*",
        default=None,
        help="""Commands to run on startup.
        If more than 1, separate commands with a semicolon (;).""",
    )

    parser.add_argument(
        "-t",
        "--trace_level",
        dest="trace_level",
        type=int,
        default=3,
        help="""Set the trace level
        0: Error only
        1: Warning/Error
        2: Info/Warning/Error
        3: All messages
        Default: 3""",
    )

    args = parser.parse_args()

    hci = BleHci(
        args.serial_port,
        baud=args.baudRate,
        id_tag=args.idtag,
        async_callback=print,
        evt_callback=print,
        flowcontrol=args.enable_flow_control,
        recover_on_power_loss=True,
    )
    hci.logger.setLevel(args.trace_level)

    print("Bluetooth Low Energy HCI tool")
    print(f"Serial port: {args.serial_port}")
    print(f"Monitor Trace Msg Serial Port: {args.monPort}")
    print(f"8N1 {args.baudRate}")

    commands = args.commands
    if commands:
        if len(commands) > 1:
            commands = " ".join(commands)
        else:
            commands = commands[0]

        commands = commands.split(";")
        commands = [x.strip() for x in commands]

        print("Startup commands: ")
        for command in commands:
            print(f"\t{command}")

    # Start the terminal argparse
    terminal = ArgumentParser(prog="", add_help=True)
    subparsers = terminal.add_subparsers()

    clear_parser = subparsers.add_parser(
        "clear",
        aliases=["cls"],
        help="Clear the screen",
        formatter_class=RawTextHelpFormatter,
    )

    clear_parser.set_defaults(
        func=lambda _: os.system("cls" if os.name == "nt" else "clear")
    )
    #### UPDATE PARSER ####
    update_parser = subparsers.add_parser(
        "update", help="update the firmware", formatter_class=RawTextHelpFormatter
    )
    update_parser.add_argument("update", help="name of application file")
    update_parser.set_defaults(
        func=lambda args: print(hci.firmware_update(args.update)),
        which="update",
    )

    #### RESET PARSER ####
    reset_parser = subparsers.add_parser(
        "sysreset", help="reset the firmware", formatter_class=RawTextHelpFormatter
    )

    reset_parser.set_defaults(
        func=lambda args: print(hci.reset_device()),
        which="sysreset",
    )

    #### SET_FLASH_ADDR PARSER ####
    set_addr_parser = subparsers.add_parser(
        "setflash",
        help="set the flash start address",
        formatter_class=RawTextHelpFormatter,
    )

    set_addr_parser.add_argument("addr", help="start address of memory bank to upload")

    set_addr_parser.set_defaults(
        func=lambda args: print(hci.set_flash_addr(args.addr)),
        which="setflash",
    )

    #### ERASE PARSER ####
    erase_parser = subparsers.add_parser(
        "erase", help="erase the flash", formatter_class=RawTextHelpFormatter
    )

    erase_parser.set_defaults(
        func=lambda args: print(hci.erase_memory()),
        which="erase",
    )

    #### ADDR PARSER ####
    addr_parser = subparsers.add_parser(
        "addr", help="Set the device address.", formatter_class=RawTextHelpFormatter
    )
    addr_parser.add_argument("addr", help="Device address, ex: 00:11:22:33:44:55")
    addr_parser.set_defaults(
        func=lambda args: print(hci.set_address(args.addr)),
        which="addr",
    )

    memstats_parser = subparsers.add_parser(
        "memstats",
        help="Get BLE stack memory usage statistics",
        formatter_class=RawTextHelpFormatter,
    )

    memstats_parser.set_defaults(func=lambda _: print(hci.get_memory_stats()))

    #### ADV PARSER ####
    adv_parser = subparsers.add_parser(
        "adv-start",
        help="Start advertising",
        formatter_class=RawTextHelpFormatter,
    )
    adv_parser.add_argument(
        "-i",
        "--interval",
        dest="adv_interval",
        type=_hex_int,
        default=DEFAULT_ADV_INTERVAL,
        help=f"""Advertising interval in units of 0.625 ms, 16-bit hex number 0x0020 - 0x4000.
        Default: 0x{DEFAULT_ADV_INTERVAL}""",
    )
    adv_parser.add_argument(
        "-n", "--name", type=str, default="", help="Advertising name"
    )
    adv_parser.add_argument(
        "--no-connect",
        dest="connect",
        action="store_false",
        help="Disable advertising as a connectable device.",
    )

    adv_parser.set_defaults(
        func=lambda args: print(
            hci.start_advertising(
                connect=args.connect,
                adv_params=AdvParams(
                    adv_type=0 if args.connect else 0x3,
                    interval_min=args.adv_interval,
                    interval_max=args.adv_interval,
                ),
                adv_name=args.name,
            )
        ),
        which="adv",
    )
    adv_stop_parser = subparsers.add_parser(
        "adv-stop",
        help="Stop advertising",
        formatter_class=RawTextHelpFormatter,
    )
    adv_stop_parser.set_defaults(func=lambda _: print(hci.enable_adv(False)))

    #### SCAN PARSER ####
    scan_parser = subparsers.add_parser(
        "scan",
        help="Send scanning commands and print scan reports, ctrl-c to exit.",
        formatter_class=RawTextHelpFormatter,
    )
    scan_parser.add_argument(
        "-i",
        "--interval",
        dest="scan_interval",
        type=_hex_int,
        default=DEFAULT_SCAN_INTERVAL,
        help=f"""Scanning interval in units of 0.625 ms, 16-bit hex number 0x0004 - 0x4000.
        Default: 0x{DEFAULT_SCAN_INTERVAL}""",
    )
    scan_parser.set_defaults(
        func=lambda args: print(
            hci.set_scan_params(
                scan_params=ScanParams(scan_interval=args.scan_interval)
            )
        )
        or hci.enable_scanning(enable=True),
        which="scan",
    )

    #### INIT PARSER ####
    init_parser = subparsers.add_parser(
        "init",
        help="Send the initiating commands to open a connection",
        formatter_class=RawTextHelpFormatter,
    )
    init_parser.add_argument(
        "addr", help="Address of peer to connect with, ex: 00:11:22:33:44:55"
    )
    init_parser.add_argument(
        "-i",
        "--interval",
        dest="conn_interval",
        type=_hex_int,
        default=DEFAULT_CONN_INTERVAL,
        help=f"""Connection interval in units of 1.25ms, 16-bit hex number 0x0006 - 0x0C80."
        Default: {hex(DEFAULT_CONN_INTERVAL)}""",
    )
    init_parser.add_argument(
        "-t",
        "--timeout",
        dest="sup_timeout",
        type=_hex_int,
        default=DEFAULT_SUP_TIMEOUT,
        help=f"""Supervision timeout in units of 10ms, 16-bit hex number 0x000A - 0x0C80.
        "Default: {hex(DEFAULT_SUP_TIMEOUT)}""",
    )
    init_parser.set_defaults(
        func=lambda args: print(
            hci.init_connection(
                addr=convert_str_address(args.addr[::-1]),
                interval=args.conn_interval,
                sup_timeout=args.sup_timeout,
                conn_params=EstablishConnParams(
                    peer_addr=convert_str_address(args.addr)
                ),
            )
        )
    )

    datalen_parser = subparsers.add_parser(
        "data-len", help="Set the max data length", formatter_class=RawTextHelpFormatter
    )
    datalen_parser.set_defaults(func=lambda _: hci.set_data_len(), which="dataLen")

    send_acl_parser = subparsers.add_parser(
        "send-acl",
        help="Send ACL packets",
        formatter_class=RawTextHelpFormatter,
    )
    send_acl_parser.add_argument(
        "packet_length",
        type=int,
        help="Number of bytes per ACL packet, 16-bit decimal 1-65535, 0 to disable.",
    )
    send_acl_parser.add_argument(
        "num_packets",
        type=int,
        help="Number of packets to send, 8-bit decimal 1-255, 0 to enable auto-generate",
    )
    send_acl_parser.add_argument(
        "--handle",
        type=int,
        default=0,
        help="Number of bytes per ACL packet, 16-bit decimal 1-65535, 0 to disable.",
    )
    send_acl_parser.set_defaults(
        func=lambda args: print(
            hci.generate_acl(args.handle, args.packet_len, args.num_packets)
        ),
    )

    sinl_acl_parser = subparsers.add_parser(
        "sink-acl",
        help="Sink ACL packets, do not send events to host",
        formatter_class=RawTextHelpFormatter,
    )
    sinl_acl_parser.add_argument("-e", "--enable", default=1)
    sinl_acl_parser.set_defaults(
        func=lambda _: print(hci.enable_acl_sink(bool(args.enable))),
    )
    adv_stats_parser = subparsers.add_parser(
        "adv-stats",
        aliases=["as"],
        help="Get the advertising stats",
        formatter_class=RawTextHelpFormatter,
    )

    adv_stats_parser.set_defaults(func=lambda _: print(hci.get_adv_stats()))

    scan_stats_parser = subparsers.add_parser(
        "scan-stats",
        aliases=["ss"],
        help="Get the scan stats",
        formatter_class=RawTextHelpFormatter,
    )

    scan_stats_parser.set_defaults(func=lambda _: print(hci.get_scan_stats()))

    conn_stats_parser = subparsers.add_parser(
        "conn-stats",
        aliases=["cs"],
        help="Get the connection stats",
        formatter_class=RawTextHelpFormatter,
    )

    conn_stats_parser.set_defaults(
        func=lambda _: print(hci.get_conn_stats()), which="connstats"
    )

    test_stats_parser = subparsers.add_parser(
        "test-stats",
        aliases=["ts"],
        help="Get the test stats",
        formatter_class=RawTextHelpFormatter,
    )

    test_stats_parser.set_defaults(func=lambda _: print(hci.get_test_stats()))

    rssi_parser = subparsers.add_parser(
        "rssi",
        help="Get an RSSI sample using CCA",
        formatter_class=RawTextHelpFormatter,
    )
    rssi_parser.add_argument("-c", "--channel", default=0)

    def _print_rssi(_args):
        rssi, status = hci.get_rssi_vs(_args.channel)
        print(f"RSSI (dBm): {rssi}")
        print(status)

    rssi_parser.set_defaults(func=_print_rssi)

    #### RESET PARSER ####
    reset_parser = subparsers.add_parser("reset", help="Sends an HCI reset command")
    reset_parser.set_defaults(func=lambda _: print(hci.reset()), which="reset")

    #### TX TEST PARSER ####
    tx_test_parser = subparsers.add_parser(
        "tx-test",
        aliases=["tx"],
        help="Execute the transmitter test.",
        formatter_class=RawTextHelpFormatter,
    )
    tx_test_parser.add_argument(
        "-c",
        "--channel",
        type=int,
        dest="channel",
        default=0,
        help="TX test channel, 0-39. Default: 0",
    )
    tx_test_parser.add_argument(
        "--phy",
        dest="phy",
        type=int,
        default=1,
        help="""Tx Test PHY
        1: 1M
        2: 2M
        3: S8
        4: S2
        Default: 1M""",
    )
    tx_test_parser.add_argument(
        "-p",
        "--payload",
        dest="payload",
        type=int,
        default=0,
        help="""Tx Test Payload
        0: PRBS9
        1:11110000
        2:10101010
        3: PRBS15
        4: 11111111
        5:00000000
        6:00001111
        7: 01010101
        Default: PRBS9""",
    )
    tx_test_parser.add_argument(
        "-pl",
        "--packet-length",
        type=int,
        default=0,
        help="Tx packet length, number of bytes per packet, 0-255. Default: 0",
    )

    tx_test_parser.set_defaults(
        func=lambda args: print(
            hci.tx_test(
                channel=args.channel,
                phy=PhyOption(args.phy),
                payload=PayloadOption(args.payload),
                packet_len=args.packet_length,
            )
        ),
    )

    tx_test_vs_parser = subparsers.add_parser(
        "txtestvs",
        aliases=["txvs"],
        help="Execute the vendor-specific transmitter test",
        formatter_class=RawTextHelpFormatter,
    )
    tx_test_vs_parser.add_argument(
        "-c",
        "--channel",
        type=int,
        dest="channel",
        default=0,
        help="Tx test channel. Default: 0",
    )
    tx_test_vs_parser.add_argument(
        "--phy",
        dest="phy",
        type=int,
        default=1,
        help="""Tx Test PHY
        1: 1M
        2: 2M
        3: S8
        4:S2
        Default: 1M""",
    )
    tx_test_vs_parser.add_argument(
        "-p",
        "--payload",
        dest="payload",
        type=int,
        default=0,
        help="""Tx Test Payload
        0: PRBS9
        1: 11110000
        2: 10101010
        3: PRBS15
        4: 11111111
        5: 00000000
        6: 00001111
        7: 01010101
        Default: PRBS9
        """,
    )
    tx_test_vs_parser.add_argument(
        "-pl",
        "--packet-length",
        dest="packet_length",
        type=int,
        default=0,
        help="Tx packet length, number of bytes per packet, 0-255. Default: 0",
    )
    tx_test_vs_parser.add_argument(
        "-n",
        "--num-packets",
        dest="num_packets",
        type=int,
        default=0,
        help="Number of packets to transmit, 2 bytes hex, 0 equals infinite. Default: 0",
    )
    tx_test_vs_parser.set_defaults(
        func=lambda args: print(
            hci.tx_test_vs(
                channel=args.channel,
                phy=args.phy,
                payload=args.payload,
                packet_len=args.packet_length,
                num_packets=args.num_packets,
            )
        )
    )

    #### RXTEST PARSER ####
    rx_test_parser = subparsers.add_parser(
        "rx-test",
        aliases=["rx"],
        help="Execute the receiver test",
        formatter_class=RawTextHelpFormatter,
    )
    rx_test_parser.add_argument(
        "-c",
        "--channel",
        dest="channel",
        type=int,
        default=0,
        help="Rx test channel, 0-39. Default: 0",
    )
    rx_test_parser.add_argument(
        "--phy",
        dest="phy",
        type=int,
        default=1,
        help="""Rx Test PHY
        1: 1M
        2: 2M
        3: S8
        4: S2
        Default: 1""",
    )
    rx_test_parser.set_defaults(
        func=lambda args: print(
            hci.rx_test(channel=args.channel, phy=PhyOption(args.phy))
        )
    )

    #### RXTESTVS PARSER ####
    rx_test_vs_parser = subparsers.add_parser(
        "rx-test-vs",
        aliases=["rxvs"],
        help="Execute the vendor-specific receiver test",
        formatter_class=RawTextHelpFormatter,
    )
    rx_test_vs_parser.add_argument(
        "-c",
        "--channel",
        dest="channel",
        type=int,
        default=0,
        help="Rx test channel. Default: 0",
    )
    rx_test_vs_parser.add_argument(
        "--phy",
        dest="phy",
        type=int,
        default=1,
        help="""Rx Test PHY
        1: 1M
        2: 2M
        3: S8
        4: S2
        Default: 1""",
    )
    rx_test_vs_parser.add_argument(
        "-m",
        "--mod-idx",
        dest="modulationIdx",
        type=int,
        default=0,
        help="Vendor-specific modulation index",
    )
    rx_test_vs_parser.add_argument(
        "-n",
        "--num_packets",
        dest="num_packets",
        type=int,
        default=0,
        help="Number of packets to transmit, 2 bytes hex, 0 equals infinite. Default: 0",
    )
    rx_test_vs_parser.set_defaults(
        func=lambda args: print(
            hci.rx_test_vs(
                channel=args.channel,
                phy=PhyOption(args.phy),
                modulation_idx=args.modulationIdx,
                num_packets=args.num_packets,
            )
        ),
    )

    #### ENDTEST PARSER ####
    endtest_parser = subparsers.add_parser(
        "end-test",
        aliases=["end"],
        help="End the Tx/Rx test, print the number of correctly received packets",
        formatter_class=RawTextHelpFormatter,
    )

    def _end_test(_args):
        rx_packets, status = hci.end_test()
        print(f"RX Received: {rx_packets}")
        print(status)

    endtest_parser.set_defaults(func=_end_test)

    #### RESET TEST STATS PARSER ####
    reset_test_stats_parser = subparsers.add_parser(
        "reset-ts",
        aliases=["rsts"],
        help="Reset accumulated stats from DTM",
        formatter_class=RawTextHelpFormatter,
    )
    reset_test_stats_parser.set_defaults(
        func=lambda _: print(hci.reset_test_stats()), which="reset-test-stats"
    )

    #### RESET CONNECTION STATS PARSER ####
    reset_connection_stats_parser = subparsers.add_parser(
        "reset-cs",
        aliases=["rscs"],
        help="Reset accumulated stats from connection mode",
        formatter_class=RawTextHelpFormatter,
    )
    reset_connection_stats_parser.set_defaults(
        func=lambda _: print(hci.reset_connection_stats()),
        which="reset-connection-stats",
    )
    #### RESET Adv STATS PARSER ####
    reset_adv_stats_parser = subparsers.add_parser(
        "reset-adv-stats",
        aliases=["rsas"],
        help="Reset accumulated stats from connection mode",
        formatter_class=RawTextHelpFormatter,
    )
    reset_adv_stats_parser.set_defaults(
        func=lambda _: print(hci.reset_adv_stats()),
    )

    reset_scan_stats_parser = subparsers.add_parser(
        "reset-scan-stats",
        aliases=["rsss"],
        help="Reset accumulated stats from connection mode",
        formatter_class=RawTextHelpFormatter,
    )
    reset_scan_stats_parser.set_defaults(
        func=lambda _: print(hci.reset_scan_stats()),
    )

    #### SET PHY PARSER ####
    set_phy_parser = subparsers.add_parser(
        "set-phy",
        aliases=["sp"],
        help="Set the PHY.",
        formatter_class=RawTextHelpFormatter,
    )
    set_phy_parser.add_argument(
        "-c",
        "--conn-handle",
        type=int,
        dest="handle",
        default=0,
        help="Connection handle. Default: 0",
    )
    set_phy_parser.add_argument(
        "-p",
        "--phy",
        dest="phy",
        type=int,
        default=1,
        help="""Desired PHY
        1: 1M
        2: 2M
        3: S2
        4: S8
        Default: 1M""",
    )
    phy_lut = {
        1: PhyOption.PHY_1M,
        2: PhyOption.PHY_2M,
        3: PhyOption.PHY_CODED_S2,
        4: PhyOption.PHY_CODED_S8,
    }
    set_phy_parser.set_defaults(
        func=lambda _: print(
            hci.set_phy(
                handle=args.handle, tx_phys=phy_lut[args.phy], rx_phys=phy_lut[args.phy]
            )
        ),
        which="set-phy",
    )

    #### TXPOWER PARSER ####
    txpower_parser = subparsers.add_parser(
        "tx-power",
        aliases=["txp"],
        help="Set the Tx power",
        formatter_class=RawTextHelpFormatter,
    )
    txpower_parser.add_argument(
        "power", type=int, help="Integer power setting in units of dBm."
    )

    txpower_parser.set_defaults(
        func=lambda args: print(hci.set_adv_tx_power(args.power)),
    )

    discon_parser = subparsers.add_parser(
        "discon",
        aliases=["dc"],
        help="Send the command to disconnect",
        formatter_class=RawTextHelpFormatter,
    )

    discon_parser.set_defaults(func=lambda _: print(hci.disconnect()), which="discon")

    set_ch_map_parser = subparsers.add_parser(
        "set-chmap",
        help="Set the connection channel map to a given channel.",
        formatter_class=RawTextHelpFormatter,
    )
    set_ch_map_parser.add_argument(
        "channel",
        type=int,
        nargs="?",
        help="""Channel to use in channel map
        Will set the channel map to the given channel plus one additional channel.""",
    )
    set_ch_map_parser.add_argument(
        "-m",
        "--mask",
        dest="mask",
        type=_hex_int,
        help=""""4-byte hex number to use a channel map.
        0xFFFFFFFFFF uses all channels, 0x000000000F will use channels 0-3, etc""",
    )
    set_ch_map_parser.add_argument(
        "--handle",
        dest="handle",
        type=int,
        default=0,
        help="Connection handle, integer. Default: 0",
    )
    set_ch_map_parser.set_defaults(
        func=lambda args: hci.set_channel_map(channels=args.mask, handle=args.handle),
    )

    cmd_parser = subparsers.add_parser(
        "cmd", help="Send raw HCI command", formatter_class=RawTextHelpFormatter
    )
    cmd_parser.add_argument(
        "cmd",
        help='String of hex bytes LSB first\nex: "01030C00" to send HCI Reset command',
    )
    cmd_parser.add_argument(
        "-t",
        "--timeout",
        dest="timeout",
        type=int,
        help="Command timeout, Default: None",
    )

    cmd_parser.set_defaults(
        func=lambda args: print(hci.write_command_raw(bytes.fromhex(args.command)))
    )

    #### EXIT PARSER ####
    exit_parser = subparsers.add_parser(
        "exit",
        aliases=["quit", "q"],
        help="Exit the program",
        formatter_class=RawTextHelpFormatter,
    )
    exit_parser.set_defaults(func=lambda _: sys.exit(EXIT_FUNC_MAGIC), which="exit")

    ls_parser = subparsers.add_parser(
        "ls",
        help="List directory",
        formatter_class=RawTextHelpFormatter,
    )

    ls_parser.add_argument("ls_dir", nargs="?", default=".")
    ls_parser.set_defaults(
        func=lambda args: [print(x) for x in os.listdir(args.ls_dir)]
    )
    cd_parser = subparsers.add_parser(
        "cd",
        help="change working directory",
        formatter_class=RawTextHelpFormatter,
    )
    cd_parser.add_argument("dir")
    cd_parser.set_defaults(func=lambda args: os.chdir(args.dir))

    pwd_parser = subparsers.add_parser(
        "pwd",
        help="print working directory",
        formatter_class=RawTextHelpFormatter,
    )
    pwd_parser.set_defaults(func=lambda args: print(os.getcwd()))

    run_parser = subparsers.add_parser(
        "run",
        help="run command via os",
        formatter_class=RawTextHelpFormatter,
    )
    run_parser.add_argument("run", nargs="+")
    run_parser.set_defaults(func=lambda args: os.system(" ".join(args.run)))

    #### HELP PARSER ####
    help_parser = subparsers.add_parser("help", aliases=["h"], help="Show help message")
    help_parser.set_defaults(func=lambda _: terminal.print_help(), which="help")

    def _completer(text, state):
        commands = subparsers.choices.keys()
        matches = [cmd for cmd in commands if cmd.startswith(text)]
        return matches[state] if state < len(matches) else None

    readline.set_completer(_completer)
    readline.parse_and_bind("tab: complete")
    readline.set_completer_delims(readline.get_completer_delims().replace("-", ""))

    command_run = False
    if commands:
        command_run = _run_input_cmds(commands, terminal)

    while True:
        if commands and not command_run:
            logger.info("Port set, running input commands.")
            command_run = _run_input_cmds(commands, terminal)

        command_str = input(f"{command_state}>>> ")

        # just an empty command
        if command_str in ("", os.linesep):
            continue

        try:
            args = terminal.parse_args(command_str.split())
            try:
                args.func(args)
            except AttributeError as err:
                logger.error(str(err))
                continue

        except SystemExit as err:
            if err.code == EXIT_FUNC_MAGIC:
                sys.exit(0)
            elif err.code != 0:
                logger.error(
                    "Process finished with exit code %s (%s)", err, type(err).__name__
                )

        except Exception as err:  # pylint: disable=broad-exception-caught
            logger.error("Unexpected exception %s", type(err).__name__)


################## MAIN ##################
if __name__ == "__main__":
    main()



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""DOCSTRING"""
from dataclasses import dataclass
from enum import Enum


class PacketType(Enum):
    """BT standard packet types."""

    COMMAND = 0x1
    """Command packet type."""

    ASYNC = 0x2
    """Asynchronous Connection-Less packet type."""

    SYNC = 0x3
    """Synchronous packet type.
    
    .. note::
        Synchronous data packets are not used in BLE.
    
    """

    EVENT = 0x4
    """Event packet type."""

    EXTENDED = 0x9
    """Extended command packet type."""


class OGF(Enum):
    """BLE-defined Opcode Group Field values."""

    NOP = 0x00
    """No operation."""

    LINK_CONTROL = 0x01
    """Link control group field."""

    LINK_POLICY = 0x02
    """Link policy group field."""

    CONTROLLER = 0x03
    """Controller group field."""

    INFORMATIONAL = 0x04
    """Informational group field."""

    STATUS = 0x05
    """Status group field."""

    TESTING = 0x06
    """Testing group field."""

    LE_CONTROLLER = 0x08
    """LE controller group field."""

    VENDOR_SPEC = 0x3F
    """Vendor specific group field."""


class NOpOCF(Enum):
    """BLE-defined NOP group Opcode Command Field values."""

    NOP = 0x00
    """No operation."""


class LinkControlOCF(Enum):
    """BLE-defined Link Control group Opcode Command Field values"""

    DISCONNECT = 0x06
    """Disconnect command."""

    READ_REMOTE_VER_INFO = 0x1D
    """Read remote version info command."""


class ControllerOCF(Enum):
    """BLE-defined Controller group Opcode Command Field values."""

    SET_EVENT_MASK = 0x01
    """Set event mask command."""

    RESET = 0x03
    """Reset command."""

    READ_TX_PWR_LVL = 0x2D
    """Read TX power level command."""

    SET_CONTROLLER_TO_HOST_FC = 0x31
    """Set controller to host flow control command."""

    HOST_BUFFER_SIZE = 0x33
    """Host buffer size command."""

    HOST_NUM_CMPL_PKTS = 0x35
    """Host number of completed packets command."""

    SET_EVENT_MASK_PAGE2 = 0x63
    """Set event mask page 2 command."""

    READ_AUTH_PAYLOAD_TO = 0x7B
    """Read authenticated payload timeout command."""

    WRITE_AUTH_PAYLOAD_TO = 0x7C
    """Write authenticated payload timeout command."""

    CONFIG_DATA_PATH = 0x83
    """Configure data path command."""


class InformationalOCF(Enum):
    """BLE-defined Information group Opcode Command Field values."""

    READ_LOCAL_VER_INFO = 0x01
    """Read local version information command."""

    READ_LOCAL_SUP_CMDS = 0x02
    """Read local supported commands command."""

    READ_LOCAL_SUP_FEAT = 0x03
    """Read local supported features command."""

    READ_BUF_SIZE = 0x05
    """Read buffer size command."""

    READ_BD_ADDR = 0x09
    """Read BD address command."""

    READ_LOCAL_SUP_CODECS = 0x0D
    """Read local supported codecs command."""

    READ_LOCAL_SUP_CODEC_CAP = 0x0E
    """Read local supported codec capabilities command."""

    READ_LOCAL_SUP_CONTROLLER_DLY = 0x0F
    """Read local supported controller delay command."""


class StatusOCF(Enum):
    """BLE-defined Status group Opcode Command Field values."""

    READ_RSSI = 0x05
    """Read RSSI command."""


class LEControllerOCF(Enum):
    """BLE-defined LE Controller group Opcode Command Field values."""

    SET_EVENT_MASK = 0x01
    """Set event mask command."""

    READ_BUF_SIZE = 0x02
    """Read buffer size command."""

    READ_LOCAL_SUP_FEAT = 0x03
    """Read local supported features command."""

    SET_RAND_ADDR = 0x05
    """Set random address command."""

    SET_ADV_PARAM = 0x06
    """Set advertising parameters command."""

    READ_ADV_TX_POWER = 0x07
    """Read advertising physical channel TX power command."""

    SET_ADV_DATA = 0x08
    """Set advertising data command."""

    SET_SCAN_RESP_DATA = 0x09
    """Set scan response data command."""

    SET_ADV_ENABLE = 0x0A
    """Set advertising enable command."""

    SET_SCAN_PARAM = 0x0B
    """Set scan parameters command."""

    SET_SCAN_ENABLE = 0x0C
    """Set scan enable command."""

    CREATE_CONN = 0x0D
    """Create connection command."""

    CREATE_CONN_CANCEL = 0x0E
    """Create connection cancel command."""

    READ_WHITE_LIST_SIZE = 0x0F
    """Read fileter accept list size command."""

    CLEAR_WHITE_LIST = 0x10
    """Clear filter accept list command."""

    ADD_DEV_WHITE_LIST = 0x11
    """Add device to filter accept list command."""

    REMOVE_DEV_WHITE_LIST = 0x12
    """Remove device from filter accept list command."""

    CONN_UPDATE = 0x13
    """Connection update command."""

    SET_HOST_CHAN_CLASS = 0x14
    """Set host channel classification command."""

    READ_CHAN_MAP = 0x15
    """Read channel map command."""

    READ_REMOTE_FEAT = 0x16
    """Read remove features command."""

    ENCRYPT = 0x17
    """Encrypt command."""

    RAND = 0x18
    """Rand command."""

    START_ENCRYPTION = 0x19
    """Enable encryption command."""

    LTK_REQ_REPL = 0x1A
    """Long term key requency reply command."""

    LTK_REQ_NEG_REPL = 0x1B
    """Long term key request negative reply command."""

    READ_SUP_STATES = 0x1C
    """Read supported states command."""

    RECEIVER_TEST = 0x1D
    """Receiver test command."""

    TRANSMITTER_TEST = 0x1E
    """Transmitter test command."""

    TEST_END = 0x1F
    """Test end command."""

    REM_CONN_PARAM_REP = 0x20
    """Remote connection parameter request reply command."""

    REM_CONN_PARAM_NEG_REP = 0x21
    """Remote connection parameter request negative reply command."""

    SET_DATA_LEN = 0x22
    """Set data length command."""

    READ_DEF_DATA_LEN = 0x23
    """Read suggested default data length command."""

    WRITE_DEF_DATA_LEN = 0x24
    """Write suggested default data length command."""

    READ_LOCAL_P256_PUB_KEY = 0x25
    """Read local P-256 public key command."""

    GENERATE_DHKEY = 0x26
    """Generate DHKey command."""

    ADD_DEV_RES_LIST = 0x27
    """Add device to resolving list command."""

    REMOVE_DEV_RES_LIST = 0x28
    """Remove device from resolving list command."""

    CLEAR_RES_LIST = 0x29
    """Clear resolving list command."""

    READ_RES_LIST_SIZE = 0x2A
    """Read resolving list size command."""

    READ_PEER_RES_ADDR = 0x2B
    """Read peer resolvable address command."""

    READ_LOCAL_RES_ADDR = 0x2C
    """Read local resolvable address command."""

    SET_ADDR_RES_ENABLE = 0x2D
    """Set address resolution enable command."""

    SET_RES_PRIV_ADDR_TO = 0x2E
    """Set resolvable private address timeout command."""

    READ_MAX_DATA_LEN = 0x2F
    """Read maximum data length command."""

    READ_PHY = 0x30
    """Read PHY command."""

    SET_DEF_PHY = 0x31
    """Set default PHY command."""

    SET_PHY = 0x32
    """Set PHY command."""

    ENHANCED_RECEIVER_TEST = 0x33
    """Enhanced receiver test command."""

    ENHANCED_TRANSMITTER_TEST = 0x34
    """Enhanced transmitter test command."""

    SET_ADV_SET_RAND_ADDR = 0x35
    """Set advertising set random address command."""

    SET_EXT_ADV_PARAM = 0x36
    """Set extended advertising parameters command."""

    SET_EXT_ADV_DATA = 0x37
    """Set extended advertising data command."""

    SET_EXT_SCAN_RESP_DATA = 0x38
    """Set extended scan response data command."""

    SET_EXT_ADV_ENABLE = 0x39
    """Set extended advertising enable command."""

    READ_MAX_ADV_DATA_LEN = 0x3A
    """Read maximum advertising data length command."""

    READ_NUM_SUP_ADV_SETS = 0x3B
    """Read number of supported advertising sets command."""

    REMOVE_ADV_SET = 0x3C
    """Remove advertising set command."""

    CLEAR_ADV_SETS = 0x3D
    """Clear advertising sets command."""

    SET_PER_ADV_PARAM = 0x3E
    """Set periodic advertising parameters command."""

    SET_PER_ADV_DATA = 0x3F
    """Set periodic advertising data command."""

    SET_PER_ADV_ENABLE = 0x40
    """Set periodic advertising enable command."""

    SET_EXT_SCAN_PARAM = 0x41
    """Set extended scan parameters command."""

    SET_EXT_SCAN_ENABLE = 0x42
    """Set extended scan enable command."""

    EXT_CREATE_CONN = 0x43
    """Extended create connection command."""

    PER_ADV_CREATE_SYNC = 0x44
    """Periodic advertising create sync command."""

    PER_ADV_CREATE_SYNC_CANCEL = 0x45
    """Periodic advertising create sync cancel command."""

    PER_ADV_TERM_SYNC = 0x46
    """Periodic advertising terminate sync command."""

    ADD_DEV_PER_ADV_LIST = 0x47
    """Add devoce to periodic advertiser list command."""

    REMOVE_DEV_PER_ADV_LIST = 0x48
    """Remove device from periodic advertiser list command."""

    CLEAR_PER_ADV_LIST = 0x49
    """Clear periodic advertiser list command."""

    READ_PER_ADV_LIST_SIZE = 0x4A
    """Read periodic advertising list size command."""

    READ_TX_POWER = 0x4B
    """Read TX power command."""

    READ_RF_PATH_COMP = 0x4C
    """Read RF path compensation command."""

    WRITE_RF_PATH_COMP = 0x4D
    """Write RF path compensation command."""

    SET_PRIVACY_MODE = 0x4E
    """Set privacy mode command."""

    RECEIVER_TEST_V3 = 0x4F
    """Receiver test version3 command."""

    TRANSMITTER_TEST_V3 = 0x50
    """Transmitter test version3 command."""

    SET_CONNLESS_CTE_TX_PARAMS = 0x51
    """Set connectionless CTE transmit parameters command."""

    SET_CONNLESS_CTE_TX_ENABLE = 0x52
    """Set connectionless CTE transmit enable command."""

    SET_CONNLESS_IQ_SAMP_ENABLE = 0x53
    """Set connectionless IQ sampling enable command."""

    SET_CONN_CTE_RX_PARAMS = 0x54
    """Set connection CTE receive parameters command."""

    SET_CONN_CTE_TX_PARAMS = 0x55
    """Set connection CTE transmit parameters command."""

    CONN_CTE_REQ_ENABLE = 0x56
    """Connection CTE request enable command."""

    CONN_CTE_RSP_ENABLE = 0x57
    """Connection CTE response enable command."""

    READ_ANTENNA_INFO = 0x58
    """Read antenna information command."""

    SET_PER_ADV_RCV_ENABLE = 0x59
    """Set periodic advertising receive enable command."""

    PER_ADV_SYNC_TRANSFER = 0x5A
    """Periodic advertising sync transfer command."""

    PER_ADV_SET_INFO_TRANSFER = 0x5B
    """Periodic advertising set info transfer command."""

    SET_PAST_PARAM = 0x5C
    """Set periodic advertising sync transfer parameters command."""

    SET_DEFAULT_PAST_PARAM = 0x5D
    """Set default periodic advertising sync transfer parameters command."""

    GENERATE_DHKEY_V2 = 0x5E
    """Generate DHKey version2 command."""

    MODIFY_SLEEP_CLK_ACC = 0x5F
    """Modify sleep clock accuracy command."""

    READ_BUF_SIZE_V2 = 0x60
    """Read buffer size version2 command."""

    READ_ISO_TX_SYNC = 0x61
    """Read ISO TX sync command."""

    SET_CIG_PARAMS = 0x62
    """Set CIG parameters command."""

    SET_CIG_PARAMS_TEST = 0x63
    """Set CIG parameters test command."""

    CREATE_CIS = 0x64
    """Create CIS command."""

    REMOVE_CIG = 0x65
    """Remove CIG command."""

    ACCEPT_CIS_REQ = 0x66
    """Accept CIS request command."""

    REJECT_CIS_REQ = 0x67
    """Reject CIS request command."""

    CREATE_BIG = 0x68
    """Create BIG command."""

    CREATE_BIG_TEST = 0x69
    """Create BIG test command."""

    TERMINATE_BIG = 0x6A
    """Terminate BIG command."""

    BIG_CREATE_SYNC = 0x6B
    """BIG create sync command."""

    BIG_TERMINATE_SYNC = 0x6C
    """BIG terminate sync command."""

    REQUEST_PEER_SCA = 0x6D
    """Request peer SCA command."""

    SETUP_ISO_DATA_PATH = 0x6E
    """Setup ISO data path command."""

    REMOVE_ISO_DATA_PATH = 0x6F
    """Remove ISO data path command."""

    ISO_TX_TEST = 0x70
    """ISO TX test command."""

    ISO_RX_TEST = 0x71
    """ISO RX test command."""

    ISO_READ_TEST_COUNTERS = 0x72
    """ISO read test counters command."""

    ISO_TEST_END = 0x73
    """ISO test end command."""

    SET_HOST_FEATURE = 0x74
    """Set host feature command."""

    READ_ISO_LINK_QUAL = 0x75
    """Read ISO link quality command."""

    READ_ENHANCED_TX_POWER = 0x76
    """Enhanced read TX power level command."""

    READ_REMOTE_TX_POWER = 0x77
    """Read remote TX power command."""

    SET_PATH_LOSS_REPORTING_PARAMS = 0x78
    """Set path loss repoting parameters command."""

    SET_PATH_LOSS_REPORTING_ENABLE = 0x79
    """Set path loss reporting enable command."""

    SET_TX_POWER_REPORT_ENABLE = 0x7A
    """Set TX power reporting enable command."""

    SET_DATA_RELATED_ADDRESS_CHANGES = 0x7C
    """Set data related address changes command."""

    SET_DEF_SUBRATE = 0x7D
    """Set default subrate command."""

    SUBRATE_REQ = 0x7E
    """Subrate request command."""


class VendorSpecificOCF(Enum):
    """ADI Vendor Specific group Opcode Command Field values."""

    SET_FLASH_ADDR = 0x3C3
    """set flash address pointer"""

    DEVICE_RESET = 0x3C2
    """reset the device"""

    MEMORY_ERASE = 0x3C1
    """Erase the memory"""

    WRITE_FLASH = 0x3C0
    """WRITE the firmware"""

    SET_SCAN_CH_MAP = 0x3E0
    """Set scan channel map command."""

    SET_EVENT_MASK = 0x3E1
    """Set event mask command."""

    ENA_ACL_SINK = 0x3E3
    """Enable ACL sink command."""

    GENERATE_ACL = 0x3E4
    """Generate ACL command."""

    ENA_AUTO_GEN_ACL = 0x3E5
    """Enable autogenerate ACL command."""

    SET_TX_TEST_ERR_PATT = 0x3E6
    """Set TX test error pattern command."""

    SET_CONN_OP_FLAGS = 0x3E7
    """Set connection operational flags command."""

    SET_P256_PRIV_KEY = 0x3E8
    """Set P-256 private key command."""

    GET_PER_CHAN_MAP = 0x3DE
    """Get periodic scan/advertising channel map command."""

    GET_ACL_TEST_REPORT = 0x3E9
    """Get ACL test report command."""

    SET_LOCAL_MIN_USED_CHAN = 0x3EA
    """Set local minimum number of used channels command."""

    GET_PEER_MIN_USED_CHAN = 0x3EB
    """Get peer minimum number of used channels command."""

    VALIDATE_PUB_KEY_MODE = 0x3EC
    """Set validate public key mode command."""

    SET_BD_ADDR = 0x3F0
    """Set BD address command."""

    GET_RAND_ADDR = 0x3F1
    """Get random address command."""

    SET_LOCAL_FEAT = 0x3F2
    """Set local features command."""

    SET_OP_FLAGS = 0x3F3
    """Set operational flags command."""

    SET_ADV_TX_PWR = 0x3F5
    """Set advertising TX power command."""

    SET_CONN_TX_PWR = 0x3F6
    """Set connection TX power command."""

    SET_ENC_MODE = 0x3F7
    """Set encryption mode command."""

    SET_CHAN_MAP = 0x3F8
    """Set channel map command."""

    SET_DIAG_MODE = 0x3F9
    """Set diagnostic mode command."""

    SET_SNIFFER_ENABLE = 0x3CD
    """Set sniffer packet forwarding enable command."""

    GET_PDU_FILT_STATS = 0x3F4
    """Get PDU filter statistics command."""

    GET_SYS_STATS = 0x3FA
    """Get system statistics command."""

    GET_ADV_STATS = 0x3FB
    """Get advertising statistics command."""

    GET_SCAN_STATS = 0x3FC
    """Get scan statistics command."""

    GET_CONN_STATS = 0x3FD
    """Get connection statistics command."""

    GET_TEST_STATS = 0x3FE
    """Get test statistics command."""

    GET_POOL_STATS = 0x3FF
    """Get pool statistics command."""

    SET_AUX_DELAY = 0x3D0
    """Set auxiliary packet offset delay command."""

    SET_EXT_ADV_FRAG_LEN = 0x3D1
    """Set extended advertising data fragmentation length command."""

    SET_EXT_ADV_PHY_OPTS = 0x3D2
    """Set extended advertising PHY options command."""

    SET_EXT_ADV_DEF_PHY_OPTS = 0x3D3
    """Set extended advertising default PHY options command."""

    GENERATE_ISO = 0x3D5
    """Generate ISO packets command."""

    GET_ISO_TEST_REPORT = 0x3D6
    """Get ISO test report command."""

    ENA_ISO_SINK = 0x3D7
    """Enable ISO sink command."""

    ENA_AUTO_GEN_ISO = 0x3D8
    """Enable autogenerate ISO packets command."""

    GET_CIS_STATS = 0x3D9
    """Get CIS statistics command."""

    GET_AUX_ADV_STATS = 0x3DA
    """Get auxiliary advertising statistics command."""

    GET_AUX_SCAN_STATS = 0x3DB
    """Get auxiliary scan statistics command."""

    GET_PER_SCAN_STATS = 0x3DC
    """Get periodic scan statistics command."""

    SET_CONN_PHY_TX_PWR = 0x3DD
    """Set connection PHY TX power command."""

    REG_WRITE = 0x300
    """Register write command."""

    REG_READ = 0x301
    """Register read command."""

    RESET_CONN_STATS = 0x302
    """Reset connection statistics command."""

    TX_TEST = 0x303
    """Transmitter test command."""

    RESET_TEST_STATS = 0x304
    """Reset test statistics command."""

    RX_TEST = 0x305
    """Receiver test command."""

    GET_RSSI = 0x306
    """Get RSSI command."""

    RESET_ADV_STATS = 0x307
    """PHY enable command."""

    RESET_SCAN_STATS = 0x308
    """PHY disable command."""


@dataclass
class OCF:
    """Supported Opcode Command Field values."""

    NOP = NOpOCF
    """NOP ground OCF values."""

    LINK_CONTROL = LinkControlOCF
    """Link Control group OCF values."""

    LINK_POLICY = None
    """Link Policy group OCF values."""

    CONTROLLER = ControllerOCF
    """Controller group OCF values."""

    INFORMATIONAL = InformationalOCF
    """Informational group OCF values."""

    STATUS = StatusOCF
    """Status group OCF values."""

    TESTING = None
    """Testing group OCF values."""

    LE_CONTROLLER = LEControllerOCF
    """LE Controller group OCF values."""

    VENDOR_SPEC = VendorSpecificOCF
    """ADI Vendor Specific group OCF values."""



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""
integration.py

Description: Integration tests for HCI

"""
import os
import secrets
import sys
import unittest


from max_ble_hci import BleHci
from max_ble_hci import packet_codes as pc
from max_ble_hci.constants import PhyOption, PubKeyValidateMode


if len(sys.argv) > 1:
    PORT = sys.argv[1]
    # gotta get rid of it before unittest takes it
    sys.argv.pop()
elif os.environ.get("TEST_PORT"):
    PORT = os.environ.get("TEST_PORT")
else:
    PORT = "/dev/serial/by-id/usb-FTDI_FT230X_Basic_UART_DT03NOCL-if00-port0"

hci1 = BleHci(PORT, id_tag="hci1", timeout=5)

MAX_U32 = 0xFFFFFFFF
A32 = 0xAAAAAAAA


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
class TestHci(unittest.TestCase):
    def test_reset(self):
        # Reset puts code into nice condition, make sure it works before any tests
        self.assertEqual(hci1.reset(), pc.StatusCode.SUCCESS)

    def test_commands(self):
        hci1.reset()

        self.assertEqual(hci1.set_tx_test_err_pattern(A32), pc.StatusCode.SUCCESS)
        self.assertIsNotNone(hci1.set_connection_op_flags(1, MAX_U32, True))

        key = list(secrets.token_bytes(32))
        self.assertEqual(hci1.set_256_priv_key(key), pc.StatusCode.SUCCESS)
        self.assertEqual(
            hci1.get_channel_map_periodic_scan_adv(1, False)[1],
            pc.StatusCode.SUCCESS,
        )
        self.assertIsNotNone(hci1.get_acl_test_report())

        self.assertEqual(
            hci1.set_local_num_min_used_channels(PhyOption.PHY_1M, 0, 10),
            pc.StatusCode.SUCCESS,
        )

        self.assertEqual(
            hci1.get_peer_min_num_channels_used(1)[1],
            pc.StatusCode.ERROR_CODE_UNKNOWN_CONN_ID,
        )

        self.assertEqual(
            hci1.set_validate_pub_key_mode(PubKeyValidateMode.ALT1),
            pc.StatusCode.SUCCESS,
        )

        addr, status = hci1.get_rand_address()

        self.assertTrue(addr > 0 or status != pc.StatusCode.SUCCESS)

        self.assertEqual(hci1.set_local_feature(0), pc.StatusCode.SUCCESS)

        self.assertEqual(hci1.set_operational_flags(0, True), pc.StatusCode.SUCCESS)

        self.assertEqual(
            hci1.set_encryption_mode(1, True, True),
            pc.StatusCode.ERROR_CODE_UNKNOWN_CONN_ID,
        )

        self.assertEqual(hci1.set_diagnostic_mode(True), pc.StatusCode.SUCCESS)

    def test_stats(self):
        """Test status functions"""
        stats, status = hci1.get_pdu_filter_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)

        stats, status = hci1.get_memory_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)

        stats, status = hci1.get_adv_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)
        stats, status = hci1.get_scan_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)

        stats, status = hci1.get_conn_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)

        stats, status = hci1.get_test_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)

        stats, status = hci1.get_aux_adv_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)

        stats, status = hci1.get_aux_scan_stats()
        self.assertTrue(stats is not None and status == pc.StatusCode.SUCCESS)
        self.assertEqual(
            hci1.set_connection_phy_tx_power(1, 0, PhyOption.PHY_1M),
            pc.StatusCode.ERROR_CODE_UNKNOWN_CONN_ID,
        )


if __name__ == "__main__":
    unittest.main()



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""Contains HCI constants definitions."""
from __future__ import annotations

from enum import Enum
from typing import Tuple

ADI_PORT_BAUD_RATE = 115200

MAX_U16 = 2**16 - 1
"""Maximum value for a 16-bit unsigned integer."""
MAX_U32 = 2**32 - 1
"""Maximum value for a 32-bit unsigned integer."""
MAX_U64 = 2**64 - 1
"""Maximum value for a 64-bit unsigned integer."""


class Endian(Enum):
    """Endian byte-order definitions."""

    LITTLE = "little"
    """Little endian byte order."""

    BIG = "big"
    """Big endian byte order."""


class PhyOption(Enum):
    """BLE-defined PHY options."""

    PHY_1M = 0x1
    """1M PHY option."""

    PHY_2M = 0x2
    """2M PHY option."""

    PHY_CODED = 0x3
    """Generic coded PHY option."""

    PHY_CODED_S8 = 0x3
    """Coded S8 PHY option."""

    PHY_CODED_S2 = 0x4
    """Coded S2 PHY option."""

    def __str__(self) -> str:
        if self == PhyOption.PHY_1M:
            return "1M"

        if self == PhyOption.PHY_2M:
            return "2M"

        if self == PhyOption.PHY_CODED_S2:
            return "S2"

        if self == PhyOption.PHY_CODED_S8:
            return "S8"

        if self == PhyOption.PHY_CODED:
            return "Coded"

        return ""

    @staticmethod
    def to_mask(phy: PhyOption) -> Tuple[int, int]:
        """Convert list of phy options to PHY preference mask and PHY option mask

        Parameters
        ----------
        phy : PhyOption
            PhyOption Enum

        Returns
        -------
        Tuple[int, int]
            PHY preference mask as defined by BLE spec
            PHY option mask as defined by BLE spec
        """

        if phy == PhyOption.PHY_1M:
            return (1 << 0, 0)

        if phy == PhyOption.PHY_2M:
            return (1 << 1, 0)

        if phy in (PhyOption.PHY_CODED, PhyOption.PHY_CODED_S8):
            return (1 << 2, 2)

        return (1 << 2, 1)

    @staticmethod
    def str_to_enum(phy: str) -> PhyOption:
        """Convert phy as string to enum

        Parameters
        ----------
        phy : str
            PHY (1M, 2M, S8, S2)

        Returns
        -------
        PhyOption
            PHY option as enum

        Raises
        ------
        ValueError
            If PHY option does not exist
        """
        if phy.upper() not in ("1M", "2M", "S8", "S2"):
            raise ValueError("PHY string must be in (1M, 2M, S8, S2)")

        if phy == "1M":
            return PhyOption.PHY_1M
        if phy == "2M":
            return PhyOption.PHY_2M
        if phy == "S2":
            return PhyOption.PHY_CODED_S2
        if phy == "S8":
            return PhyOption.PHY_CODED_S8

        return PhyOption.PHY_1M


class PayloadOption(Enum):
    """BLE-definded payload options."""

    PLD_PRBS9 = 0
    """PRBS9 payload option."""

    PLD_11110000 = 1
    """11110000 payload option."""

    PLD_10101010 = 2
    """10101010 payload option."""

    PLD_PRBS15 = 3
    """PRBS15 payload option."""

    PLD_11111111 = 4
    """11111111 payload option."""

    PLD_00000000 = 5
    """00000000 payload option."""

    PLD_00001111 = 6
    """00001111 payload option."""

    PLD_01010101 = 7
    """01010101 payload option."""


class AddrType(Enum):
    """BLE-defined peer address types."""

    PUBLIC = 0
    """Public device address."""

    RANDOM = 1
    """Random device address."""

    PUBLIC_IDENTITY = 2
    """
    Resolvable Private address based on local IRK from
    resolving list, or public address if no match is found.
    
    .. note::
        For advertising, this value is valid for own device
        address type only. For connection, this value is valid
        for both own device address type and connectable peer
        device address type.

    """

    RANDOM_IDENTITY = 3
    """
    Resolvable Private address based on local IRK from
    resolving list, or random address is no match is found.

    .. note::
        For advertising, this value is valid for own device
        address type only. For connection, this value is valid
        for both own device address type and connectable peer
        device address type.

    """


class PubKeyValidateMode(Enum):
    """Public key validation modes."""

    ALT1 = 0x0
    """ALT1 validation mode."""

    ALT2 = 0x1
    """ALT2 validation mode."""



#! /usr/bin/env python3
###############################################################################
#
#
# Copyright (C) 2023 Maxim Integrated Products, Inc., All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL MAXIM INTEGRATED BE LIABLE FOR ANY CLAIM, DAMAGES
# OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# Except as contained in this notice, the name of Maxim Integrated
# Products, Inc. shall not be used except as stated in the Maxim Integrated
# Products, Inc. Branding Policy.
#
# The mere transfer of this software does not imply any licenses
# of trade secrets, proprietary technology, copyrights, patents,
# trademarks, maskwork rights, or any other form of intellectual
# property whatsoever. Maxim Integrated Products, Inc. retains all
# ownership rights.
#
##############################################################################
#
# Copyright 2023 Analog Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################
"""
enc_test.py

Description: Small encryption HCI test

"""
import argparse
import sys
from datetime import datetime
import os

# pylint: disable=import-error,redefined-builtin
from cryptography.hazmat.primitives.asymmetric import ec

from rich import print


# pylint: enable=import-error,redefined-builtin
from max_ble_hci import BleHci
from max_ble_hci.hci_packets import EventPacket
from max_ble_hci.packet_codes import EventSubcode, StatusCode


class Tester:
    """Test Harness"""

    # pylint: disable=too-few-public-methods
    P256_P = (
        115792089210356248762697446949407573530086143415290314195533631308867097853951
    )
    P256_A = -3
    P256_B = (
        41058363725152142129326129780047268409114441015993725554835256314039467401291
    )

    def __init__(self, serial_port) -> None:
        self.serial_port = serial_port
        self.hci = BleHci(
            serial_port, id_tag="hci1", timeout=5, evt_callback=self._common_callback
        )

        self.event_done = False
        self.results = {}

    def _is_point_on_curve(self, xcoord, ycoord) -> bool:
        left = ycoord * ycoord
        right = (xcoord * xcoord * xcoord) + (self.P256_A * xcoord) + self.P256_B
        return (left - right) % self.P256_P == 0

    def _check_point(self, xcoord, ycoord):
        xcoord = xcoord % self.P256_P
        ycoord = ycoord % self.P256_P

        if not self._is_point_on_curve(xcoord, ycoord):
            return False

        return True

    def _pub_key_read_event(self, packet: EventPacket):
        packet.get_return_params()

        status, xcoord, ycoord = packet.get_return_params([1, 32, 32])
        status = StatusCode(status)

        if status != StatusCode.SUCCESS:
            print(f"Read pub key failed. Status: {status}")

        self.results["pub-key-read"] = self._check_point(xcoord, ycoord)

    def _dhk_event(self, packet: EventPacket):
        if (
            StatusCode(packet.evt_params[0])
            == StatusCode.ERROR_CODE_INVALID_HCI_CMD_PARAMS
        ):
            self.results["dhk"] = True
        else:
            self.results["dhk"] = False

    def _common_callback(self, packet: EventPacket):
        if packet.evt_subcode == EventSubcode.READ_LOCAL_P256_PUB_KEY_CMPLT:
            self._pub_key_read_event(packet)
        elif packet.evt_subcode == EventSubcode.GENERATE_DHKEY_CMPLT:
            self._dhk_event(packet)

        self.event_done = True

    def _run_pub_key_read(self):
        self.event_done = False
        status = self.hci.read_local_p256_pub_key()

        if status != StatusCode.SUCCESS:
            self.results["pub-key-read"] = False
            return

        while not self.event_done:
            pass

    def _run_bad_dhk(self):
        """
        TEST HCI/AEN/BI-01-C
        Generate DH Key Error With Invalid Point
        """
        # Generate a private key for use with the P-256 curve

        private_key = ec.generate_private_key(ec.SECP256R1())

        # Generate the corresponding public key
        public_key = private_key.public_key()

        # Get the x and y coordinates from the public key
        public_numbers = public_key.public_numbers()
        invalid_x_coordinate = public_numbers.x + 1_000_000
        y_coordinate = public_numbers.y

        self.event_done = False
        start = datetime.now()
        status = self.hci.generate_dhk(xcoord=invalid_x_coordinate, ycoord=y_coordinate)

        alt1_maybe_fail = False
        if status == StatusCode.SUCCESS:
            alt1_maybe_fail = True
        elif status == StatusCode.ERROR_CODE_INVALID_HCI_CMD_PARAMS:
            self.results["dhk"] = True
            self.event_done = True
            return

        while not self.event_done and (datetime.now() - start).total_seconds() < 10:
            continue

        if not self.event_done and alt1_maybe_fail:
            self.results["dhk"] = False

    def _run_simple_aes(self):
        expected_ct = "66C6C2278E3B8E053E7EA326521BAD99"
        expected_ct = list(bytes.fromhex(expected_ct))

        evt = self.hci.write_command_raw(
            "01172020bf01fb9d4ef3bc36d874f5394138684c1302f1e0dfcebdac7968574635241302"
        )

        if evt.status != StatusCode.SUCCESS:
            self.results["encryption"] = False
        else:
            cyphertext = list(evt.evt_params[4:])

            if cyphertext != expected_ct:
                self.results["encryption"] = False
            else:
                self.results["encryption"] = True

        if not self.results["encryption"]:
            print("Failed basic AES test")
            print("Resp", cyphertext)
            print("Expected", expected_ct)

    def run(self) -> bool:
        """RUN Encryption test

        Returns
        -------
        bool
            True if all pass. False if any failure
        """
        status = self.hci.reset()
        if status != StatusCode.SUCCESS:
            return False

        status = self.hci.enable_all_events()
        if status != StatusCode.SUCCESS:
            print("Failed to enable events")
            return False

        self._run_pub_key_read()
        self._run_bad_dhk()
        self.hci.disable_all_events()
        self._run_simple_aes()

        total_result = True
        for test, result in self.results.items():
            print(f"{test}:", "[green]Pass[/green]" if result else "[red]Fail[/red]")
            if not result:
                total_result = False

        return total_result


def main():
    """MAIN"""
    port = ""

    parser = argparse.ArgumentParser(description="Basic HCI Encryption tests")
    parser.add_argument("serial_port", nargs="?", help="Serial port used for HCI")
    args = parser.parse_args()

    if args.serial_port and args.serial_port != "":
        print(args.serial_port)
        port = args.serial_port
    elif os.environ.get("TEST_PORT"):
        port = os.environ.get("TEST_PORT")

    if port == "":
        raise ValueError(
            "Serial port must be specified through command line or in script"
        )

    tester = Tester(port)
    result = tester.run()

    if not result:
        sys.exit(-1)
    else:
        sys.exit(0)


if __name__ == "__main__":
    main()



import toml
import sys
import argparse


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Automatically update version in pytoml"
    )
    parser.add_argument("pytoml")

    parser.add_argument("--major", action="store_true")
    parser.add_argument("--minor", action="store_true")
    parser.add_argument("--patch", action="store_true")

    args = parser.parse_args()

    if args.patch or (not args.major and not args.minor):
        do_patch = True
    else:
        do_patch = False

    pytoml = toml.load(args.pytoml)

    major, minor, patch = pytoml["project"]["version"].split(".")

    print(f"Current version {major}.{minor}.{patch}")

    if args.major:
        print("Updating major")
        major = int(major) + 1
        patch = 0
        minor = 0

    if args.minor:
        print("Updating minor")
        minor = int(minor) + 1
        patch = 0

    if do_patch:
        print("Updating patch")
        patch = int(patch) + 1

    new_version = f"{major}.{minor}.{patch}"
    pytoml["project"]["version"] = new_version

    print(f"New version: {new_version}")

    with open(args.pytoml, "w") as new_toml:
        toml.dump(pytoml, new_toml)
